import os,sys,glob,json
from abacustest.lib_prepare.abacus import AbacusStru,ReadInput,WriteInput,WriteKpt
from ..model import Model
from . import comm
import numpy as np
import matplotlib.pyplot as plt
import traceback
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE
from typing import List,Tuple,Optional,Union,Literal

class Phonon(Model):
    '''
    Base class for all models
    
    Some specific workflow will be implemented in the subclass.
    Such as EOS, BAND, etc.
    
    The model should have the following methods:
    '''
    @staticmethod
    def model_name():
        '''
        Name of the model, which will be used as the subcommand
        '''
        return "phonon"
    
    @staticmethod
    def description():
        '''
        Description of the model
        '''
        return "Prepare and postprocess the phonon calculation"
    
    @staticmethod
    def add_args(parser):
        '''
        Add arguments for the model
        The arguments can not be command, model, modelcommand 
        '''
    
    def run(self,params):
        '''
        Parse the parameters and run the model
        '''
        return

    @staticmethod
    def prepare_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand '''
        parser.description = "This script is used to preapre the calculation of phonon calculation by abacus and phonopy."
        parser.add_argument('--abacus_command', type=str,  default=RECOMMAND_COMMAND,help='the command to run abacus job')
        parser.add_argument('--abacus_image', type=str,  default=RECOMMAND_IMAGE,help='the image to run abacus job')
        parser.add_argument('--abacus_machine', type=str,  default=RECOMMAND_MACHINE,help='the machine to run abacus job')
        parser.add_argument('-p', '--phonopy', type=str,  default="phonopy",help='the path of phonopy executable, default is phonopy') 
        parser.add_argument('-j', '--job',default=["."], action="extend",nargs="*" ,help='the path of abacus inputs, default is the current path.')
        parser.add_argument('-s', '--setting',default="setting.conf",help='the setting file for phonopy, default is setting.conf')
        
    
    def run_prepare(self,params):
        '''
        Parse the parameters and run the prepare process.
        Usually, this step will generate the input files for abacustest submit.
        '''
        abacus_command = params.abacus_command
        abacus_image = params.abacus_image
        abacus_machine = params.abacus_machine
        phonopy = params.phonopy
        jobs = params.job if len(params.job) == 1 else params.job[1:]
        print(params)
        
        setting = PreparePhono(jobs,abacus_command,abacus_image,abacus_machine,phonopy,params.setting)
        if setting:
            comm.doc_after_prepare("Phonon", setting["run_dft"]["example"], ["setting.json"],has_prepare=False)  
    
    @staticmethod
    def postprocess_args(parser):
        '''
        Add arguments for the postprocess subcommand
        The arguments can not be command, model, modelcommand'''
        parser.description = "This script is used to post-process the calculation of phonon calculation by abacus and phonopy."
        parser.add_argument('-p', '--phonopy', type=str,  default="phonopy",help='the path of phonopy executable, default is phonopy') 
        parser.add_argument('-j', '--job',default=["."], action="extend",nargs="*" ,help='the path of abacus inputs, default is the current path.')
        parser.add_argument('-s', '--setting',default="setting.conf",help='the setting file for phonopy, default is setting.conf')
        parser.add_argument('--onlyplot', action='store_true', help='only plot the phonon band structure, do not run the phonopy postprocess')
        parser.add_argument('--compare', action='store_true', help='compare the phonon band structure with different jobs')
        parser.add_argument('--type-label', default=None, nargs="*", help='the label of the band structure, used when compare is set')
        
    def run_postprocess(self,params):
        '''
        Parse the parameters and run the postprocess process
        '''
        phonopy = params.phonopy
        jobs = params.job if len(params.job) == 1 else params.job[1:]
        if params.compare:
            if len(jobs) < 2:
                print("Error: at least two jobs are needed to compare.")
                return
            type_label = params.type_label
            if type_label is None:
                type_label = [f"Job-{i+1}" for i in range(len(jobs))]
            elif len(type_label) != len(jobs):
                print("Error: the length of type_label should be the same as the number of jobs.")
                return
            picture_file = "phonon_band_structure_compare.png"
            plot_multiple_phonon_bands([os.path.join(i,"band.yaml") for i in jobs], picture_file, type_label)
        else:
            PostprocessPhonon(jobs,phonopy,params.setting,params.onlyplot)
    
def PreparePhono(jobs,abacus_command,abacus_image,abacus_machine,phonopy,settingf="setting.conf") :
    allexamples = []
    all_fail = True
    cwd = os.getcwd()
    for ijob in jobs:
        os.chdir(ijob)
        allfiles = os.listdir()
        if not os.path.isfile("INPUT"):
            print("Error: INPUT not exist in",ijob)
            continue
        input_param = ReadInput("INPUT")
        input_param["calculation"] = "scf"  # force set to scf
        input_param["suffix"] = "ABACUS"    # force set to ABACUS
        input_param["cal_force"] = "1"
        if not os.path.isfile("setting.conf"):
            print("Error: setting.conf not exist in",ijob)
            continue
        return_code, out, err = comm.run_command(f"{phonopy} {settingf} --abacus -d",shell=True)
        if return_code != 0:
            print(out,"\n",err)
            print("Error: run phonopy prepare failed in",ijob)
            continue
        allstrus = glob.glob("STRU-*")
        for istru in allstrus:
            idx = istru.split("-")[-1]
            os.makedirs("ABACUS-"+idx)
            os.system(f"cp -r {' '.join(allfiles)} ABACUS-{idx}")
            os.system(f"cp -r {istru} ABACUS-{idx}/STRU ")
            WriteInput(input_param,os.path.join("ABACUS-"+idx,"INPUT"))
            if os.path.isfile("phonopy_disp.yaml"):
                os.system(f"cp phonopy_disp.yaml ABACUS-{idx}/")
            allexamples.append(os.path.join(ijob,f"ABACUS-{idx}"))
        all_fail = False
        os.chdir(cwd)
    
    if all_fail:
        print("Error: all jobs are failed.")
        return {}
    setting = {
        "save_path": ".",
        "bohrium_group_name": "Phonopy",
        "run_dft": {
            "example": allexamples,
            "command": abacus_command,
            "image": abacus_image,
            "bohrium": {
                "scass_type": abacus_machine,
                "job_type": "container",
                "platform": "ali",
            },
        },
    }
    comm.dump_setting(setting)  
    return setting

def PostprocessPhonon(jobs,phonon,settingf="setting.conf",only_plot=False):
    if len(jobs) == 0:
        print("No jobs are specified.")
        return
    
    cwd = os.getcwd()
    supermetrics = {}
    for ijob in jobs:
        if not os.path.isdir(ijob):
            continue
        os.chdir(ijob)
        
        if not only_plot:
            if not os.path.isfile("phonopy_disp.yaml"):
                print("Error: phonopy_disp.yaml not exist in",ijob)
                continue
            
            os.system(f"{phonon} -f */OUT.*/running_*.log")
            #return_code, out, err = comm.run_command(f"{phonon} -f */OUT.*/running_*.log",shell=True)
            #if return_code != 0:
            #    print(out,"\n",err)
            #    print("Error: catch force failed in",ijob)
            #    continue

            os.system(f"{phonon} -p {settingf} --abacus -s")
            #return_code, out, err = comm.run_command(f"{phonon} -p {settingf} --abacus -s",shell=True)
            #if return_code != 0:
            #    print(out,"\n",err)
            #    print("Error: run phonopy postprocess failed in",ijob)
            #    continue
        
        pngfile ="phonon_band_structure.png"
        if plot_phonon("band.yaml",save_file=pngfile) is None:
            print("Error: plot phonon failed in",ijob)
            continue
        else:
            supermetrics[f"{ijob}-phonon"] = {"type":"image","file":os.path.join(ijob,pngfile)}
            print(f"Plot phonon band structure in {ijob} saved in {os.path.join(ijob,pngfile)} successfully!!!")
        os.chdir(cwd)
        
    if len(supermetrics) > 0:
        json.dump(supermetrics, open("supermetrics.json", "w"), indent=4)
    
def rearrange_labels(distances, frequencies, labels, new_labels):
    new_distances = []
    new_frequencies = []
    end_distance = 0
    for label in new_labels:
        if label in labels:
            label_idx = labels.index(label)
            reverse = False
        elif label[::-1] in labels:
            label_idx = labels.index(label[::-1])
            reverse = True
        else:
            
            raise ValueError(f"Label {label} not found in labels.")
            
        idis = distances[label_idx]
        ifreq = frequencies[label_idx]
        if reverse:
            idis = [-1*i for i in idis[::-1] ]
            ifreq = ifreq[::-1]
        new_distances.append([i + end_distance - idis[0] for i in idis])
        new_frequencies.append(ifreq)
        end_distance += idis[-1] - idis[0]
    return new_distances, new_frequencies, new_labels

def read_band_yaml(filename):
    import yaml
    with open(filename, 'r') as file:
        data = yaml.safe_load(file)

    labels = []
    for path in data['labels']:
        labels.append((path[0], path[1]))
    
    segments = data['segment_nqpoint']
    qpoints = data['phonon']
    
    distances = []
    frequencies = []
    for q in qpoints:
        distances.append(q['distance'])
        freqs = [band['frequency'] for band in q['band']]
        frequencies.append(freqs)
    
    frequencies = np.array(frequencies) # shape (nq, nband)
    
    split_indices = np.cumsum(segments)[:-1]
    
    # split distances and frequencies
    new_distances = []
    new_frequencies = []
    begin_idx = 0
    for end_idx in split_indices:
        new_distances.append(distances[begin_idx:end_idx])
        new_frequencies.append(frequencies[begin_idx:end_idx])
        begin_idx = end_idx
    new_distances.append(distances[begin_idx:])
    new_frequencies.append(frequencies[begin_idx:])

    return new_distances, new_frequencies, labels

def plot_phonon(yaml_file="band.yaml", 
                type_label = None, 
                high_symmetry_path = None,
                save_file="band.png"):
    '''Plot the phonon band structure from yaml file
    
    Args:
        yaml_file (str|list): The yaml files to plot
        type_label (list): The label of the band structure
        high_symmetry_path (list of tuples): The high symmetry path, used to rearrange the labels
        savefile (str): The file name to save the plot
    
    Returns:
        datas (list): The data of the band structure, each element is a tuple of (distances, frequencies, q_labels)
        labels (list): The labels of the band structure
        
    If you want to plot multiple yaml files, type_label should be a list of the same length as yaml_file.
    '''

    if isinstance(yaml_file, str):
        yaml_file = [ yaml_file ]

    if type_label is not None and isinstance(type_label, str):
        type_label = [type_label]
    elif type_label is None and len(yaml_file) > 1:
        type_label = [i for i in yaml_file]
    
    if type_label is not None and len(type_label) != len(yaml_file):
        print("Error: The length of label and yaml_file should be the same.")
        return None

    # check files
    datas = []
    labels = []
    for i, file in enumerate(yaml_file):
        if not os.path.isfile(file):
            print(f"Error: {file} not exist.")
            continue
        try:
            distances, frequencies, q_labels = read_band_yaml(file)
            if high_symmetry_path is not None:
                distances, frequencies, q_labels = rearrange_labels(distances, frequencies, q_labels, high_symmetry_path)
        except:
            traceback.print_exc()
            print(f"Error: read {file} failed.")
            continue
        
        datas.append((distances, frequencies, q_labels))
        if type_label is not None:
            labels.append(type_label[i])
            
    if len(datas) == 0:
        print("Error: No valid yaml file.")
        return None 

    colors = [
        "#d62728",  # 红色
        "#1f77b4",  # 蓝色
        "#2ca02c",  # 绿色
        "#ff7f0e",  # 橙色
        "#9467bd",  # 紫色
        "#8c564b",  # 棕色
        "#e377c2",  # 粉色
        "#7f7f7f",  # 灰色
        "#bcbd22",  # 黄绿色
        "#17becf"   # 青色
    ]
    if len(datas) > len(colors):
        colors = plt.cm.viridis(np.linspace(0, 1, len(datas)))
    plt.figure(figsize=(8, 6))
    for i, (distances, frequencies, q_labels) in enumerate(datas):
        for j, distance in enumerate(distances):
            freq = np.array(frequencies[j])
            for k in range(freq.shape[1]):
                if j == 0 and k==0 and type_label is not None:
                    label = labels[i]
                else:
                    label = None
                plt.plot(distance, freq[:, k], "-", lw=1, label=label, color=colors[i])

    for x in distances[1:]:
        plt.axvline(x[0], color='black', linestyle='--', lw=0.8)

    print(q_labels)
    tick_pos = [distances[0][0]]
    tick_labels = [q_labels[0][0]]
    for idx, distance in enumerate(distances):
        if idx == 0:continue
        tick_pos.append(distance[0])
        if q_labels[idx-1][-1] == q_labels[idx][0]:
            label = q_labels[idx][0]
        else:
            label = f"{q_labels[idx-1][-1]}|{q_labels[idx][0]}"
        tick_labels.append(label)
    
    tick_pos.append(distances[-1][-1])
    tick_labels.append(q_labels[-1][1])

    print(tick_pos)
    print(tick_labels)
    plt.xticks(tick_pos, tick_labels, fontsize=12)
    plt.yticks(fontsize=12)
    plt.xlim(0, distances[-1][-1])
    plt.axhline(0, color='gray', linestyle='--', lw=0.8, alpha=0.5)
    plt.grid(axis='y', linestyle='--', alpha=0.7)
    
    y_max = np.max(frequencies)
    y_min = np.min(frequencies) - 0.1 * y_max
    if type_label is not None:
        # set the maximum Y to 1.3 * y_max
        plt.ylim(y_min, y_max * 1.3)
        if len(labels) <= 3:
            ncol = len(labels)
        else:
            ncol = int(len(labels)**0.5) + 1
        plt.legend(fontsize=12, frameon=False, loc='upper center', ncol=ncol)
    
    plt.xlabel("Q-Point", fontsize=14)
    plt.ylabel("Frequency (THz)", fontsize=14)
    
    plt.tight_layout()
    plt.savefig(save_file,dpi=300)
    plt.close()
    
    return (datas, labels)

def align_distance(distance1:List[List[float]], distance2:List[List[float]]):
    """Align distance2 to distance1 by scaling and shifting, to make the start and end of each segment the same to distance1.
    
    Such as:
        distance1 = [[0, 1, 2], [2, 3, 4]]
        distance2 = [[0, 2, 4, 6, 8], [4, 6, 8]]
        return [[0, 0.5, 1, 1.5, 2], [2, 3, 4]]
    """
    assert len(distance1) == len(distance2), "The length of distance1 and distance2 should be the same."
    
    new_distance2 = []
    for i in range(len(distance1)):
        d1 = distance1[i]
        d2 = distance2[i]
        if len(d2) < 2:
            print(f"Warning: The length of distance2[{i}] is less than 2, cannot align.")
            new_distance2.append(d2)
            continue
        scale = (d1[-1] - d1[0]) / (d2[-1] - d2[0])
        shift = d1[0] - d2[0] * scale
        new_d2 = [x * scale + shift for x in d2]
        new_distance2.append(new_d2)
    return new_distance2
    
def plot_multiple_phonon_bands(yaml_files, picture_file, type_labels=None, q_points=None,
                               dpi=300
                               ):
    """Plot multiple phonon band structures from different yaml files at different temperatures.
    
    """
    if type_labels is None:
        type_labels = [f"Job-{i+1}" for i in range(len(yaml_files))]
    elif len(type_labels) != len(yaml_files):
        print("Error: the length of type_labels should be the same as the number of yaml_files.")
        return
    
    import matplotlib.pyplot as plt
    import numpy as np
    plt.figure(figsize=(8, 4))
    
    colors = [
        "#d62728",  # 红色
        "#1f77b4",  # 蓝色
        "#2ca02c",  # 绿色
        "#ff7f0e",  # 橙色
        "#9467bd",  # 紫色
        "#8c564b",  # 棕色
        "#e377c2",  # 粉色
        "#7f7f7f",  # 灰色
        "#bcbd22",  # 黄绿色
        "#17becf"   # 青色
    ]
    
    distances0, frequencies0, labels0 = read_band_yaml(yaml_files[0])
    if q_points is not None:
        distances0, frequencies0, labels0 = rearrange_labels(distances0, frequencies0, labels0, q_points)
    
    for i, distance in enumerate(distances0):
        freq = np.array(frequencies0[i])
        for j in range(freq.shape[1]):
            label = type_labels[0] if i == 0 and j==0 else None
            plt.plot(distance, freq[:, j], "-", lw=1, label=label, color=colors[0])
    
    max_freq = np.max(frequencies0)
    for t in range(1, len(yaml_files)):
        distances_, frequencies_, labels_ = read_band_yaml(yaml_files[t])
        distances_, frequencies_, labels_ = rearrange_labels(distances_, frequencies_, labels_, labels0)
        distances_ = align_distance(distances0, distances_)
        
        for i, distance in enumerate(distances_):
            freq = np.array(frequencies_[i])
            for j in range(freq.shape[1]):
                label = type_labels[t] if i == 0 and j==0 else None
                plt.plot(distance, freq[:, j], "-", lw=1, label=label, color=colors[t % len(colors)]) 
        
        max_freq = max(max_freq, np.max(frequencies_))
    
    for x in distances0[1:]:
        plt.axvline(x[0], color='black', linestyle='--', lw=0.8)
    
    tick_pos = [distances0[0][0]]
    tick_labels = [labels0[0][0]]
    for idx, distance in enumerate(distances0):
        if idx == 0:continue
        tick_pos.append(distance[0])
        if labels0[idx-1][-1] == labels0[idx][0]:
            label = labels0[idx][0]
        else:
            label = f"{labels0[idx-1][-1]}|{labels0[idx][0]}"
        tick_labels.append(label)
    
    tick_pos.append(distances0[-1][-1])
    tick_labels.append(labels0[-1][1])  
    print(tick_pos)
    
    print(tick_labels)
    plt.xticks(tick_pos, tick_labels, fontsize=12)
    plt.yticks(fontsize=12)
    plt.xlim(0, distances0[-1][-1])
    plt.axhline(0, color='gray', linestyle='--', lw=0.5, alpha=0.5)
    plt.grid(axis='y', linestyle='--', alpha=0.7)
    plt.ylim(np.min(frequencies0)-0.1*np.max(frequencies0), np.max(frequencies0)*1.3)
    if len(type_labels) <= 3:
        ncol = len(type_labels)
    else:
        ncol = int(len(type_labels)**0.5) + 1
    plt.legend(fontsize=12, frameon=False, loc='upper center', ncol=ncol)
    plt.xlabel("Q-Point", fontsize=14)
    plt.ylabel("Frequency (THz)", fontsize=14)
    plt.tight_layout()
    plt.savefig(picture_file,dpi=dpi)
    plt.close()
    print(f"Plot phonon band structures in {yaml_files} saved in {picture_file} successfully!!!")
        
    


from typing import Dict, List
import numpy as np
import re
import sys
from .. import constant


class QeInputs:
    def __init__(self,
                 param:Dict[str,any],
                 label:List[str],
                 atom_number:List[int],
                 cell:List[List[float]] ,
                 coord:List[List[float]] ,
                 pp:List[str],
                 kpt:List[any] = None,
                 cell_type:str = "alat",
                 coord_type: str = "alat",
                 kpt_type = "automatic",
                 mass:List[float] = None,
                 element:List[str] = None,  
    ):
        """QE INPUT class, the unit is Bohr
        
        Parameters
        ----------
        param : Dict[str,any]
            the input parameters of QE.
            key is the card name: control, system, electrons, ions, cell
            value is a dictionary of the card parameters.
        label : List[str]
            the label name of each type; 
        atom_number : List[int]
            the atom number of each type
        pp : List[str]
            pseudopotential file of each type
        coord : List[List[float]]
            coordinate of each atom
        cell : List[List[float]]
            cell of a, b, c 
        kpt : List[any], optional
            k point list, by default None, which means [1,1,1,0,0,0] and kpt_type is automatic.
        cell_type : str, optional
            the type of cell, by default "alat"
        coord_type : str, optional
            the type of coordinate, by default "alat"
        kpt_type : str, optional    
            the type of k point, by default "automatic"
        mass : List[float], optional
            mass of each type, by default None
        element : List[str], optional
            element name of each type, by default None
        """
 
        self.label = label
        self.atom_number = atom_number
        self.cell = cell
        self.coord = coord
        self.pp = pp
        self.kpt = kpt
        self.cell_type = cell_type
        self.coord_type = coord_type
        self.kpt_type = kpt_type
        
        self.param = {}
        comm_cards = ["control","system","electrons","ions","cell"]
        for ik,iv in param.items():
            if ik.lower() in comm_cards:
                self.param[ik.lower()] = iv if iv != None else {}
            else:
                self.param[ik] = iv if iv != None else []
        for ikey in comm_cards:
            if ikey not in self.param:
                self.param[ikey] = {}
                
        if element == None:
            print("WARNING: element is not defined, will use the first one/two letter of label as element name")
            self.element = []
            for i in label:
                while i[-1].isdigit(): i = i[:-1]
                self.element.append(i)
        else:
            self.element = element
        
        if mass == None:
            self.mass = [constant.MASS_DICT.get(i,1.0) for i in self.element]
        else:
            self.mass = mass
        
        if kpt == None:
            self.kpt = [1,1,1,0,0,0]
            self.kpt_type = "automatic"

    def get_param(self):
        return self.param
    
    def get_label(self):
        # return the label name of each atom
        label = []
        for idx,i in enumerate(self.atom_number):
            label += [self.label[idx]] * i
        return label
    
    def get_pp(self):
        return self.pp
    
    def get_mass(self):
        return self.mass
    
    def get_kpt(self):
        return self.kpt
    
    def get_kpt_type(self):
        return self.kpt_type
    
    def get_cell(self,bohr = False):
        if self.cell_type == "bohr":
            transfer_unit = 1
        elif self.cell_type == "angstrom":
            transfer_unit = constant.A2BOHR
        elif self.cell_type == "alat":
            transfer_unit = float(self.param["system"]["celldm(1)"])
        else:
            print("ERROR: cell_type is not defined")
            sys.exit(1)
        if not bohr:
            transfer_unit *= constant.BOHR2A
            
        cell = np.array(self.cell) * transfer_unit
        return cell.tolist()
    
    def get_coord(self,bohr = False, direct=False):
        '''return the coordinate matrix, in cartesian or direct type, in unit of Angstrom or Bohr.
        If direct is True, then return the direct type coordinate, and bohr will be ignored.'''
        
        if direct and self.coord_type == "crystal":
            return self.coord
        
        if not direct:
            if self.coord_type == "crystal":
                return np.array(self.coord).dot(np.array(self.get_cell(bohr=bohr))).tolist()
            else:
                if self.coord_type == "bohr":
                    transfer_unit = 1
                elif self.coord_type == "angstrom":
                    transfer_unit = constant.A2BOHR
                elif self.coord_type == "alat":
                    transfer_unit = float(self.param["system"]["celldm(1)"])
                if not bohr:
                    transfer_unit *= constant.BOHR2A
                return (np.array(self.coord) * transfer_unit).tolist()
        else:
            if self.coord_type == "crystal":
                return self.coord
            else:
                coord_bohr = self.get_coord(bohr=True,direct=False)
                cell_bohr = self.get_cell(bohr=True)
                return np.array(coord_bohr).dot(np.linalg.inv(np.array(cell_bohr))).tolist()
    
    def _str_value(self,value):
        # if value is float or int then return str(value)
        # if value is string and not ".true.", ".false." then return "'%s'" % value
        
        if isinstance(value,str):
            try:
                float(value)
                return value
            except:
                if value.lower() in [".true.",".false."]:
                    return value
                else:
                    return "'%s'" % value
        else:
            return str(value)
    
    def write(self,filename="input"):
        with open(filename,"w") as f:
            for param_key in ["control","system","electrons","ions","cell"]:
                f.write("&%s\n" % param_key.upper())
                for key,value in self.param.get(param_key,{}).items():
                    f.write("%17s = %s\n" % (key,self._str_value(value)))
                f.write("/\n\n")
            
            f.write("\nCELL_PARAMETERS (%s)\n" % self.cell_type)
            for i in self.cell:
                f.write("%17.11f %17.11f %17.11f\n" % tuple(i))
            
            f.write("\nATOMIC_SPECIES\n")
            for i,ilabel in enumerate(self.label):
                f.write("%s %f %s\n" % (ilabel,self.mass[i],self.pp[i]))
            
            f.write("\nATOMIC_POSITIONS (%s)\n" % self.coord_type)
            coord_idx = 0
            for i,ilabel in enumerate(self.label):
                for j in range(self.atom_number[i]):
                    f.write("%5s %17.11f %17.11f %17.11f\n" % tuple([ilabel] + self.coord[coord_idx]))
                    coord_idx += 1
                    
            f.write("\nK_POINTS %s\n" % self.kpt_type)
            if self.kpt_type == "automatic":
                kpoints = self.kpt[:3]
                offset = [0 if ioff == 0 else 1 for ioff in self.kpt[3:]]
                f.write("%d %d %d %d %d %d\n" % tuple(kpoints + offset))   
            elif self.kpt_type == "gamma":
                f.write("0\n")
            else:
                print("ERROR: kpt_type is not defined")
                sys.exit(1)
            
            # write special card
            for key,value in self.param.items():
                if key.lower() not in ["control","system","electrons","ions","cell"]:
                    if isinstance(value,dict):
                        f.write("\n&%s\n" % key)
                        for ik,iv in value.items():
                            f.write("%17s = %s\n" % (ik,str(iv)))
                        f.write("/\n\n")
                    elif isinstance(value,list):
                        f.write("\n%s\n" % key)
                        for iv in value:
                            f.write("%s\n" % str(iv))
                        f.write("\n")
                    else:
                        print("ERROR: special card '%s' is not defined" % key)
                        sys.exit(1)
            
        
    @staticmethod
    def ReadInput(filename="input"):
        def get_card_type(line):
            line = re.split("[#!]",line)[0]
            if "(" in line:
                return line.split("(")[1].split(")")[0].strip().lower()
            else:
                return line.split()[1].lower()
            
        with open(filename) as f:
            lines = f.readlines()
        param = {}
        label = []
        atom_number = []
        cell = []
        coord = []
        pp = []
        kpt = None
        cell_type = "alat"
        coord_type = "alat"
        kpt_type = "automatic"
        mass = []
        element = []
        for i, line in enumerate(lines):
            if line.strip() == "" or line.strip()[0] in ["!", "#"]: continue
            elif line.strip()[0] == "&":
                block = line.split()[0].strip()[1:].lower()
                if block in ["control","system","electrons","ions","cell"]:
                    param[block] = {}
                    for ij in range(i+1,len(lines)):
                        if lines[ij].strip() == "" or lines[ij].strip()[0] in ["!","#"]: continue
                        elif lines[ij].strip() == "/": break
                        else:
                            for iline in re.split("[#!]",lines[ij])[0].split(","):
                                if iline.strip() == "":continue
                                sline = iline.split("=")
                                param[block][sline[0].strip()] = sline[1].strip()
            elif line.split("#")[0].split()[0] == "CELL_PARAMETERS":
                cell_type = get_card_type(line)
                for ij in range(i+1,len(lines)):
                    if lines[ij].strip() == "" or lines[ij].strip()[0] in ["!","#"]: continue
                    else:
                        cell.append([float(i) for i in lines[ij].split()[:3]])
                        if len(cell) == 3: break
            elif line.split("#")[0].split()[0] == "ATOMIC_SPECIES":
                for ij in range(i+1,len(lines)):
                    if lines[ij].strip() == "" or lines[ij].strip()[0] in ["!","#"]: continue
                    else:
                        sline = lines[ij].split("#")[0].split()
                        label.append(sline[0].strip())
                        atom_number.append(0)
                        mass.append(float(sline[1].strip()))
                        pp.append(sline[2].strip())
                        if len(label) == int(param["system"]["ntyp"]): break
            elif line.split("#")[0].split()[0] == "ATOMIC_POSITIONS":
                coord_type = get_card_type(line)
                label_coord = {}
                for ij in range(i+1,len(lines)):
                    if lines[ij].strip() == "" or lines[ij].strip()[0] in ["!","#"]: continue
                    elif lines[ij].strip() in ["crystal","alat","bohr","angstrom"]: continue
                    else:
                        sline = lines[ij].split("#")[0].split()
                        ilabel = sline[0].strip()
                        if ilabel not in label:
                            print("ERROR: label '%s' is not matched that in ATOMIC_SPECIES" % ilabel)
                            sys.exit(1)
                        atom_number[label.index(ilabel)] += 1
                        if ilabel not in label_coord:
                            label_coord[ilabel] = []
                        label_coord[ilabel].append([float(i) for i in sline[1:4]])
                        if sum(atom_number) == int(param["system"]["nat"]): break
                for ilabel in label:
                    coord += label_coord[ilabel]
            elif line.split("#")[0].split()[0] == "K_POINTS":
                kpt_type = get_card_type(line)
                if kpt_type == "automatic":
                    kpt = [int(i) if idx < 3 else float(i) for idx,i in enumerate(lines[lines.index(line)+1].split()[:6])]
                elif kpt_type == "crystal":
                    kpt = [int(i) for i in lines[lines.index(line)+1].split()]
                elif kpt_type == "gamma":
                    kpt = [0]
                else:
                    print("ERROR: kpt_type is not defined")
                    sys.exit(1)
        return QeInputs(param=param,
                        label=label,
                        atom_number=atom_number,
                        cell=cell,
                        coord=coord,
                        pp=pp,
                        kpt=kpt,
                        cell_type=cell_type,
                        coord_type=coord_type,
                        kpt_type=kpt_type,
                        mass=mass,
                        element=element)



import json
import os
import subprocess
import glob
import shutil
import select
import copy
#from dp.metadata import MetadataContext
#from dp.metadata.utils.storage import TiefblueStorageClient
from dflow import download_artifact, S3Artifact, config, s3_config
from dflow.plugins import bohrium
from dflow.plugins.bohrium import TiefblueClient


def create_path(output_path):
    work_path = "abacustest"  # os.path.join(output_path,"abacustest")
    # os.path.join(output_path,"abacustest_download")
    download_path = "abacustest_download"
    for ipath in [output_path, work_path, download_path]:
        os.makedirs(ipath, exist_ok=True)
    return {"output_path": os.path.abspath(output_path),
            "work_path": os.path.abspath(work_path),
            "download_path": os.path.abspath(download_path)}


def register_dflow(private_set):
    config["host"] = private_set.get("dflow_host", "").strip()
    s3_config["endpoint"] = private_set.get(
        "dflow_s3_config_endpoint", "").strip()
    config["k8s_api_server"] = private_set.get(
        "dflow_k8s_api_server", "").strip()
    config["token"] = private_set.get("dflow_token", "").strip()
    bohrium.config["username"] = private_set.get('bohrium_username', '')
   # bohrium.config["password"] = private_set.get('bohrium_password','')
    bohrium.config["ticket"] = private_set.get('bohrium_ticket', '')
    bohrium.config["project_id"] = private_set.get('bohrium_project_id', '')
    s3_config["repo_key"] = "oss-bohrium"
    s3_config["storage_client"] = TiefblueClient()


def read_config(opts):
    # parse config
    CONFIG_KEYS = ["bohrium_username", "bohrium_password", "bohrium_ticket", "bohrium_project_id",
                   "dflow_host", "dflow_s3_config_endpoint", "dflow_k8s_api_server", "dflow_token",
                   "aim_access_token", "dflow_labels", "github_username", "github_email", "github_token"]
    my_config = {}
    for ikey in CONFIG_KEYS:
        config_key = "Config_" + ikey
        if hasattr(opts, config_key):
            if ikey == "dflow_labels":
                my_config[ikey] = getattr(opts, config_key)
            elif getattr(opts, config_key).strip():
                my_config[ikey] = getattr(opts, config_key).strip()
    register_dflow(my_config)
    return my_config

def run_command(
        cmd,
        shell=True
):
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=shell,
        executable='/bin/bash'
    )
    out = ""
    err = ""
    while True:
        # 监视stdout和stderr文件描述符的可读状态
        readable, _, _ = select.select(
            [process.stdout, process.stderr], [], [])

        # 读取已经准备好的输出
        for fd in readable:
            if fd == process.stdout:
                line = process.stdout.readline()
                print(line.decode()[:-1])
                out += line.decode()
            elif fd == process.stderr:
                line = process.stderr.readline()
                print("STDERR:", line.decode()[:-1])
                err += line.decode()

        # 如果子进程已经结束，则退出循环
        return_code = process.poll()
        if return_code is not None:
            break
    return return_code, out, err


def exec_abacustest(allparams, work_path, command="abacustest submit -p param.json"):
    # work_path: to run abacustest
    # ouput_path: the write the report files
    # write param.json
    params = copy.deepcopy(allparams)
    '''
    if "config" in params:
        for ik,iv in params["config"].items():
            os.environ[ik.upper()] = str(iv)
        del params["config"]
    '''
    json.dump(params,
              open(os.path.join(work_path, "param.json"), 'w'),
              indent=4)

    cwd = os.getcwd()
    os.chdir(work_path)

    print("Execute abacustest ...")
    '''
    return_code, stdout, stderr = run_command(command)
    if return_code != 0:
        print("Error Encountered!")
        #print(stdout)
        print(stderr)
    else:
        #print(stdout)
        print(stderr)
    '''
    import abacustest.abacustest as abatest
    import argparse
    from abacustest.main import print_head
    print_head()
    parser = argparse.ArgumentParser(description="abacustest")
    subparser = parser.add_subparsers(dest="command")
    abatest.AbacusTestArgs(subparser.add_parser("submit"))
    parser = parser.parse_args(["submit", "-p", "param.json"])
    abatest.abacustest(parser)

    os.chdir(cwd)
    stdout, stderr = "", ""
    return stdout, stderr


def produce_html_table(table):
    content = "<table border=\"2px\">"
    for i, it in enumerate(table):
        if i == 0:
            content += "<thead>"
        else:
            content += "<tbody>"

        content += "<tr>"
        if len(it) == 1:
            content += f"<td colspan=\"{len(table[0])}\">" + \
                str(it[0]) + "</td>"
        else:
            for j in it:
                content += "<td>" + str(j) + "</td>"
        content += "</tr>"

        if i == 0:
            content += "</thead>"
        else:
            content += "</tbody>"
    content += "</table>"
    return content

'''
def get_datahub_dataset(bohrium_username, bohrium_password, bohrium_project, urn, download_path=None):
    metadata_storage_client = TiefblueStorageClient(
        bohrium_username, bohrium_password, bohrium_project)
    with MetadataContext(storage_client=metadata_storage_client) as context:
        dataset = context.client.get_dataset(urn)
        # print(dataset,dataset.uri)
        if dataset != None and download_path != None:
            # context.client.download_dataset(dataset,download_path)
            artifact = S3Artifact(key=dataset.uri)
            download_artifact(artifact, path=download_path)
        return dataset
'''

def unpack(filepath, output_path, filetype=None, get_support_filetype=False):
    # if file type is not supportted, just copy the file to output_path
    if get_support_filetype:
        return ["zip", "tar", "gz", "bz2", "tgz"]

    import zipfile
    import tarfile
    import gzip
    import bz2

    if not os.path.isfile(filepath):
        raise Exception("File %s not exists!" % filepath)

    if not os.path.isdir(output_path):
        os.makedirs(output_path, exist_ok=True)

    if filetype == None:
        if filepath.endswith(".tar.gz"):
            filetype = "tgz"
        else:
            filetype = os.path.splitext(filepath)[1][1:]

    if filetype == "zip":
        with zipfile.ZipFile(filepath, 'r') as zip_ref:
            zip_ref.extractall(output_path)
    elif filetype == "tar":
        with tarfile.open(filepath, "r") as tar_ref:
            tar_ref.extractall(output_path)
    elif filetype == "gz":
        with gzip.open(filepath, 'rb') as gz_ref:
            with open(output_path, 'wb') as out_ref:
                out_ref.write(gz_ref.read())
    elif filetype == "bz2":
        with bz2.open(filepath, 'rb') as bz2_ref:
            with open(output_path, 'wb') as out_ref:
                out_ref.write(bz2_ref.read())
    elif filetype == "tgz":
        with tarfile.open(filepath, "r:gz") as tar_ref:
            tar_ref.extractall(output_path)
    else:
        print("File type \'%s\' is not supported!" % filetype)
        filename = os.path.split(filepath)[-1]
        if os.path.isfile(os.path.join(output_path, filename)):
            os.remove(os.path.join(output_path, filename))
        shutil.copyfile(filepath, os.path.join(output_path, filename))

    print("Unpack %s to %s" % (filepath, output_path))
    return output_path


def pack(packfile_list, packfile_name, pack_type="zip"):
    if pack_type == "zip":
        import zipfile
        with zipfile.ZipFile(packfile_name, 'w') as zip_ref:
            for ifile in packfile_list:
                if os.path.isfile(ifile):
                    zip_ref.write(ifile)
                else:
                    for root, __, ifile in os.walk(ifile):
                        for i in ifile:
                            zip_ref.write(os.path.join(root, i))
    elif pack_type == "tar":
        import tarfile
        with tarfile.open(packfile_name, "w") as tar_ref:
            for ifile in packfile_list:
                tar_ref.add(ifile)
    elif pack_type == "gz":
        import gzip
        with gzip.open(packfile_name, 'wb') as gz_ref:
            for ifile in packfile_list:
                with open(ifile, 'rb') as in_ref:
                    gz_ref.write(in_ref.read())
    elif pack_type == "bz2":
        import bz2
        with bz2.open(packfile_name, 'wb') as bz2_ref:
            for ifile in packfile_list:
                with open(ifile, 'rb') as in_ref:
                    bz2_ref.write(in_ref.read())
    elif pack_type == "tgz":
        import tarfile
        with tarfile.open(packfile_name, "w:gz") as tar_ref:
            for ifile in packfile_list:
                tar_ref.add(ifile)
    else:
        print("Pack type \'%s\' is not supported!" % pack_type)
        return None
    return packfile_name


def download_url(url, output_path="./"):
    import requests
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        if not os.path.isdir(output_path):
            os.makedirs(output_path, exist_ok=True)
        filename = os.path.join(output_path, url.split("/")[-1])
        with open(filename, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
        return filename
    else:
        print(f"dwonload ({url}) failed, status code:", response.status_code)
        return None


def clean_dictorys(ipath):
    for ifile in glob.glob(os.path.join(ipath, "*")):
        if os.path.isdir(ifile):
            shutil.rmtree(ifile)
        else:
            os.remove(ifile)


def move_results_to_output(work_path, output_path, result_folder):
    target_result_folder = os.path.join(output_path, result_folder)
    if os.path.abspath(target_result_folder) != os.path.abspath(output_path):
        n = 1
        while os.path.isdir(target_result_folder):
            target_result_folder = os.path.join(
                output_path, result_folder + "_" + str(n))
            n += 1
    
    src_folder = os.path.join(work_path, result_folder)
    if os.path.abspath(src_folder) != os.path.abspath(work_path):
        shutil.move(os.path.join(work_path, result_folder), target_result_folder)
    else:
        # move all files in work_path to target_result_folder
        for ifile in glob.glob(os.path.join(work_path, "*")):
            shutil.move(ifile, target_result_folder)

def pack_results(output_path,result_path):
    cwd = os.getcwd()
    os.chdir(os.path.join(output_path,result_path))
    allfiles = os.listdir(".")
    alldirs = []
    allfile = []
    for f in allfiles:
        if os.path.isdir(f):
            alldirs.append(f)
        elif os.path.isfile(f) and not f.endswith(".zip") and not f.startswith("."):
            allfile.append(f)
    packed_file_name = "results.zip"
    pack(alldirs,packed_file_name,"zip")
    pack(allfile,"files.zip","zip")
    os.chdir(cwd)

def gen_dir(dir1):
    dir1 = dir1.rstrip("/")
    n = 0
    while os.path.isdir(dir1):
        dir1 = dir1 + "_" + str(n)
        n += 1
    return dir1

def capture_output(func, *args, **kwargs):
    """ Capture the output of a function and return it as a string.
    Args:
        func (callable): The function whose output is to be captured.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.
    Returns:
        str: The captured output of the function.
    """
    import sys
    from io import StringIO
    if not callable(func):
        raise ValueError("The provided func must be callable.")
    
    old_stdout = sys.stdout
    redirected_output = sys.stdout = StringIO()
    
    try:
        return_values = func(*args, **kwargs)
        output = redirected_output.getvalue()
    finally:
        sys.stdout = old_stdout
    
    return return_values, output
    


import traceback,json,sys,os
from dp.launching.typing.basic import BaseModel,String,Float,Int,Boolean,Set, List, Optional
from dp.launching.typing import Field, InputFilePath
from enum import Enum
from abacustest.lib_model.model_014_vasp2abacus import Vasp2Abacus


from . import (comm_class,
               comm_func,
               comm_report,
               ) 

class VASPSourceSet(BaseModel):
    VaspSource: List[InputFilePath] = Field(
                                         title="VASP任务目录压缩包（支持多个任务）[ENCUT/EDIFF等参数不会进行转换]",
                                         st_kwargs_type=comm_func.unpack(
                                             None, None, get_support_filetype=True),
                                         description="""上传VASP任务目录的压缩包，可对INCAR/POSCAR/KPOINTS转换成ABACUS相应的输入文件""",
                                         format="markdown")
    
    InputTemplate: Optional[String] = Field( title="ABACUS额外输入参数设置",
                                default="basis_type lcao\necutwfc 100",
                                description="""ABACUS的额外输入参数设置，如果不需要额外设置，请留空""",
                                format="multi-line")

def unpack_files(vaspsource, work_path):
    """
    Link the files from the VASP source set to the work path.
    """
    for source in vaspsource:
        source = source.get_full_path()
        if os.path.isfile(source):
            comm_func.unpack(source, work_path)
    
    dirs = [i for i in os.listdir(work_path) if os.path.isdir(os.path.join(work_path, i))]
    if len(dirs) == 0:
        dirs = ["."]
    
    return dirs

def write_input_template(input_template, work_path):
    """
    Write the input template to a file in the work path.
    """
    if input_template.strip():
        input_file = os.path.join(work_path, "INPUT.template")
        with open(input_file, "w") as f:
            f.write(input_template)
        return input_file
    return None        

class Vasp2AbacusModel(
    VASPSourceSet,
    comm_class.OutputSet,
    BaseModel):
    ...  

def Vasp2AbacusRunner(opts:Vasp2AbacusModel) -> int:
    try:
        logs = comm_class.myLog()

        paths = comm_func.create_path(str(opts.IO_output_path))
        work_path = paths["work_path"]
        output_path = paths["output_path"]
        
        logs.iprint("read source setting ...")
        
        vaspjobs = unpack_files(opts.VaspSource, work_path) # only folder names
        input_template = write_input_template(opts.InputTemplate, work_path)

        pwd = os.getcwd()
        os.chdir(work_path)
        
        vasp2abacus = Vasp2Abacus(vaspjobs, "/root/pporb/ABACUS-V1/pp", "/root/pporb/ABACUS-V1/orb", input_template)
        jobs, running_output = comm_func.capture_output(vasp2abacus.run)
        
        logs += running_output
        
        # pack jobs
        pack_filename = "ABACUS.zip"
        comm_func.pack(jobs, pack_filename)
        
        logs.iprint(f"VASP jobs converted to ABACUS jobs, and are packed in {pack_filename}")

        os.chdir(pwd)
        comm_report.gen_report(opts,logs,work_path,output_path,None)
        
    except:
        traceback.print_exc()
        return 1

    return 0



from pydantic import BaseModel, Field
from typing import List, Tuple, Union, Dict, Any, Optional, Literal
import numpy as np
import copy

from abacustest.constant import MASS_DICT, A2BOHR, BOHR2A, ABACUS_STRU_KEY_WORD
from abacustest.lib_prepare.comm import Cartesian2Direct, Direct2Cartesian, mag_to_angle, angle_to_mag

import traceback
import os
import sys

class AbacusAtomType(BaseModel):
    """ABACUS Atom Type class, which defines the common properties for a type of atoms.

    Args:
        label (str): Atom label.
        element (str): Element symbol.
        mass (float): Atomic mass.
        pp (str, optional): Pseudopotential file name.
        orb (str, optional): Orbital file name.
        paw (str, optional): PAW file name.
        type_mag (float, optional): Default magnetic moment for the atom type. Default is 0.0.
        natom (int, optional): Number of atoms of this type. Default is 0.
    """
    label: str
    element: Optional[str] = None
    mass: Optional[float] = None
    pp: Optional[str] = None
    orb: Optional[str] = None
    paw: Optional[str] = None
    type_mag: float = 0.0
    natom: int = 0

    # define equivalence method
    def __eq__(self, other):
        if not isinstance(other, AbacusAtomType):
            return NotImplemented
        mass1 = round(self.mass, 6) if self.mass is not None else None
        mass2 = round(other.mass, 6) if other.mass is not None else None
        mag1 = round(self.type_mag, 6)
        mag2 = round(other.type_mag, 6)
        return (self.label == other.label and
                self.element == other.element and
                mass1 == mass2 and
                self.pp == other.pp and
                self.orb == other.orb and
                self.paw == other.paw and
                mag1 == mag2)

    def _default_list(self):
        out = [self.label]
        out.append(self.element if self.element is not None else "")
        out.append(self.mass if self.mass is not None else 0.0)
        out.append(self.pp if self.pp is not None else "")
        out.append(self.orb if self.orb is not None else "")
        out.append(self.paw if self.paw is not None else "")
        out.append(self.type_mag)
        return out

    def __lt__(self, other):
        return self._default_list() < other._default_list()

class AbacusATOM(BaseModel):
    """ABACUS ATOM class.

    Args:
        label (str): Atom label.
        coord (Tuple[float, float, float] or np.ndarray): Atom coordinate in cartesian type with unit Angstrom.
        element (str, optional): Element symbol. If not provided, it will be inferred from the label.
        mass (float, optional): Atomic mass. If not provided, it will be set according to the element.
        pp (str, optional): Pseudopotential file name.
        orb (str, optional): Orbital file name.
        paw (str, optional): PAW file name.
        type_mag (float, optional): Default magnetic moment for the atom type. Default is 0.0.
        move (Tuple[bool, bool, bool], optional): Movement constraints along x, y, z directions. Default is (True, True, True).
        mag (float or Tuple[float, float, float], optional): Magnetic moment for the atom. Can be a scalar or a 3D vector.
        angle1 (float, optional): For non-collinear magnetism, the magnetic moment direction with respect to z-axis.
        angle2 (float, optional): For non-collinear magnetism, the magnetic moment direction projection on xy-plane with respect to x-axis.
        velocity (Tuple[float, float, float], optional): Velocity of the atom.
        constrain (bool or Tuple[bool, bool, bool], optional): If do delta_spin constrain on this atom.
        lambda_ (float or Tuple[float, float, float], optional): Lambda parameter for the atom. Can be a scalar or a 3D tuple.

    NOTE: 
        1. type_mag is referred as the value defined in STRU file for each atom type, such as:
            O   # atom label
            0.0  # magnetic moment for this atom type
            2    # atom number
            0 0 0  # coord of atom 1
            0.5 0.5 0.5  # coord of atom 2
          If mag is defined for the atom, then type_mag will be ignored.
        2. mag/angle1/angle2 are the values defined for each atom individually, such as below:
            0 0 0 mag 2 angle1 90 angle2 45  # for atom 1
            0.5 0.5 0.5 mag 0 0 2  # for atom 2
        3. for non-collinear case, ABACUS supports to set mag and angle1/angle2 at the same time, and the real magnetic moment should calculated as:
            mag_z = |mag| * cos(angle1)
            mag_x = |mag| * sin(angle1) * cos(angle2)
            mag_y = |mag| * sin(angle1) * sin(angle2)
            When only one of angle1/angle2 is provided, the another angle will be set the default value 0.0.
        4. When read/write STRU file, the type_mag/mag/angle1/angle2 is exactly the same as in the file. None value means not set in the file.
        5. Use atommag to get the real magnetic moment vector (mag_x, mag_y, mag_z) for noncolinear case, and mag for collinear case. If mag is None will return type_mag.
        6. Use set_atommag() method to set the mag/angle1/angle2 values for the atom.
    """
    label: str
    coord: Union[Tuple[float, float, float]]
    # optional
    element: Optional[str] = None
    mass: Optional[float] = None
    pp: Optional[str] = None
    orb: Optional[str] = None
    paw: Optional[str] = None
    type_mag: Optional[float] = 0.0
    move: Optional[Tuple[bool, bool, bool]] = (True, True, True)
    velocity: Optional[Tuple[float, float, float]] = None
    constrain: Optional[Union[bool, Tuple[bool, bool, bool]]] = None
    lambda_: Optional[Union[float, Tuple[float, float, float]]] = None

    mag: Optional[Union[float, Tuple[float, float, float]]] = None  # magnitude of magnetic moment
    angle1: Optional[float] =  None  # angle between magnetic moment and z-axis
    angle2: Optional[float] =  None  # angle between projection of magnetic moment on xy-plane and x-axis

    def __init__(self, **data):
        super().__init__(**data)

        if self.element is None:
            self.element = self._infer_element_from_label()
        if self.mass is None:
            self.mass = MASS_DICT.get(self.element, 0.0)

    def _infer_element_from_label(self) -> Optional[str]:
        """Infer the element symbol from the atom label."""
        if len(self.label) >= 2 and self.label[:2].capitalize() in MASS_DICT:
            return self.label[:2].capitalize()
        elif self.label[0].capitalize() in MASS_DICT:
            return self.label[0].capitalize()
        return None

    def __str__(self):
        c = f"Label: {self.label}, Element: {self.element}, Mass: {self.mass:.4f}\n"
        c += f"PP: {self.pp}\nORB: {self.orb}\n"
        c += f"Type mag: {self.type_mag}, mag: {self.mag}, angle1: {self.angle1}, angle2: {self.angle2}\n"
        c += f"Coord: {self.coord[0]:%12.7f} {self.coord[1]:%12.7f} {self.coord[2]:%12.7f}\n"
        c += f"Move: {self.move}, Velocity: {self.velocity}\n"
        c += f"Constrain: {self.constrain}, Lambda: {self.lambda_}\n"
        c += f"Real atomic mag/magnitude: {self.atommag} / {self.atommag_magnitude}\n"
        return c

    @property
    def atomtype(self) -> AbacusAtomType:
        """Get the AbacusAtomType object for this atom."""
        return AbacusAtomType(
            label=self.label,
            element=self.element,
            mass=self.mass,
            pp=self.pp,
            orb=self.orb,
            paw=self.paw,
            type_mag=self.type_mag,
            natom=1
        )

    @property
    def noncolinear(self) -> bool:
        """Check if the atom is non-collinear magnetic. Only when angle1 or angle2 is set, or mag is a 3D vector, the atom is non-collinear magnetic."""
        
        if self.angle1 is not None or self.angle2 is not None:
            return True
        if isinstance(self.mag, (list, tuple)) and len(self.mag) == 3:
            return True
        return False

    @property
    def atommag_magnitude(self) -> float:
        """Get the magnitude of the magnetic moment.

        Returns:
            float: Magnitude of the magnetic moment.
        """
        import numpy as np
        if self.mag is None:
            return abs(self.type_mag)
        else:
            return np.linalg.norm(self.mag)

    @property
    def atommag(self) -> Union[float, Tuple[float, float, float]]:
        """Get the atomic magnetic moment: one float for colinear case, or a tuple of three floats for non-colinear case.
        """
        import numpy as np
        if self.noncolinear:
            mag_norm = self.mag
            angle1 = angle2 = 0.0
            if self.angle1 is not None:
                angle1 = self.angle1    
            if self.angle2 is not None:
                angle2 = self.angle2
            mag_z = mag_norm * np.cos(np.radians(angle1))
            mag_x = mag_norm * np.sin(np.radians(angle1)) * np.cos(np.radians(angle2))
            mag_y = mag_norm * np.sin(np.radians(angle1)) * np.sin(np.radians(angle2))
            return (mag_x, mag_y, mag_z)
        else:
            if self.mag is None:
                return self.type_mag
            else:
                return self.mag
    
    def mag_angle(self):
        # Calculate angle of noncollinear magnetic moment
        if self.noncolinear:
            if self.angle1 is None and self.angle2 is None:
                angle1, angle2 = mag_to_angle(*self.mag)
            else:
                angle1, angle2 = 0, 0
                if self.angle1:
                    angle1 = self.angle1
                if self.angle2:
                    angle2 = self.angle2
            return (angle1, angle2)
        else:
            return (None, None)

    def set_atommag(self, mag: Union[float,Tuple[float,float,float]],
                    angle1: float = None,
                    angle2: float = None):
        """Set magnetic moment for the atom.

        Args:
            mag (float or Tuple[float,float,float]): Magnetic moment value.
            angle1 (float, optional): The angle between the magnetic moment and the z-axis in degrees.
            angle2 (float, optional): The angle between the projection of the magnetic moment on the xy-plane and the x-axis in degrees.
        
        If angle1 or angle2 is None, then this value will not write to STRU file.
        """
        if isinstance(mag, (list, tuple)):
            assert len(mag) == 3, f"Magnetic moment tuple must have three components, got {mag}."
        elif isinstance(mag, np.ndarray):
            mag = mag.tolist()
        elif isinstance(mag, (int, float)):
            pass
        else:
            raise TypeError("Magnetic moment must be float or tuple of three floats")

        self.mag = mag
        self.angle1 = angle1
        self.angle2 = angle2

    @staticmethod
    def sort(atom_list: List["AbacusATOM"],
             keep_first_order=True,
             only_label=False) -> List["AbacusATOM"]:
        """Rearrange atom list according to atomtype.

        Args:
            atom_list (List[AbacusATOM]): List of AbacusATOM objects.
            keep_first_order (bool): If True, keep the order of first appearance of atom types. Default is True.
            only_label (bool): If True, only sort according to label, ignoring other properties (element/mass/pp/orb/paw/type_mag).

        Returns:
            List[AbacusATOM]: Rearranged list of AbacusATOM objects.
            List[int]: Indices of the rearranged atoms in the original list.
        """
        key = "label" if only_label else "atomtype"
        unique_atom = []
        for i in atom_list:
            attr = getattr(i, key)
            if attr not in unique_atom:
                unique_atom.append(attr)

        if not keep_first_order:
            unique_atom.sort()
        new_atom_list = []
        index = []
        for atomtype in unique_atom:
            for i, atom in enumerate(atom_list):
                attr = getattr(atom, key)
                if attr == atomtype:
                    new_atom_list.append(atom)
                    index.append(i)
        return new_atom_list, index

    @staticmethod
    def find_uniq_atomtypes(atom_list: List['AbacusATOM']
                            ) -> List[AbacusAtomType]:
        """Return a list of unique AbacusAtomType objects for atom list.
        NOTE: It only merge the consecutive same atom types, which means there may be duplicate types in the returned list.
        If you want to get the true unique atom types, please use sort() method first.
        """
        unique_types = []
        if not atom_list:
            return unique_types
        unique_types.append(atom_list[0].atomtype)
        natom = 1
        for i in range(1, len(atom_list)):
            if atom_list[i].atomtype != atom_list[i-1].atomtype:
                unique_types[-1].natom = natom
                unique_types.append(atom_list[i].atomtype)
                natom = 1
            else:
                natom += 1
        unique_types[-1].natom = natom
        assert sum([atype.natom for atype in unique_types]) == len(atom_list), "Sum of unique atom types' natom does not equal to total number of atoms."
        return unique_types

    @staticmethod
    def same_type(atomlist: List['AbacusATOM']) -> bool:
        """Check if all atoms in the list are of the same type by comparing their label/element/mass/pp/orb/paw/type_mag.

        Args:
            atom_list (List[AbacusATOM]): List of AbacusATOM objects.
        Returns:
            bool: True if all atoms are of the same type, False otherwise.
        """
        if len(atomlist) in [0, 1]:
            return True
        
        for i in range(1, len(atomlist)):
            if atomlist[i].atomtype != atomlist[0].atomtype:
                return False
        return True
    
    def rotate(self, rot_mat: np.ndarray):
        """
        Rotate the cartesian coordinates of an atom. Related properties, including velocity will be updated.
        """
        # Rotate cartesian coordinate
        self.coord = tuple(np.dot(rot_mat, np.array(self.coord)))

        # Rotate velocity
        if self.velocity is not None:
            self.velocity = tuple(np.dot(rot_mat, np.array(self.velocity)))
        
        # Rotate non-collinear magnetic moment
        if self.noncolinear:
            new_mag = np.dot(rot_mat, np.array(self.atommag))
            if self.angle1 is None or self.angle2 is None:
                self.set_atommag(new_mag.tolist())
            else:
                angle1, angle2 = mag_to_angle(*new_mag)
                self.set_atommag(np.linalg.norm(new_mag).tolist(),
                                 angle1,
                                 angle2)


class AbacusSTRU:
    """ABACUS STRU class

    """

    def __init__(self,
                 cell: Union[List[Tuple[float,float,float]], np.ndarray],
                 atoms: List[AbacusATOM],
                 dpks: Optional[str] = None,
                 metadata: Dict[str, Any] = {}, 
                 ):
        self._cell = cell
        self._atoms = atoms
        self.dpks = dpks
        if isinstance(self.cell, np.ndarray):
            assert self.cell.ndim == 2 and self.cell.shape == (3,3), "Cell numpy array must be two-dimensional with shape (3,3)."
            self.cell = [tuple(self.cell[i].tolist()) for i in range(3)]
        self.metadata = {
            "lattice_constant": 1.0,
            "atom_type": "cartesian", # or "direct"
        }
        if metadata:
            self.metadata.update(metadata)

    def __str__(self):
        chem_labels = "".join([self.labels_uniq[i]+str(self.labels.count(self.labels_uniq[i])) for i in range(len(self.labels_uniq))])
        c = f"ABACUS STRU object:{chem_labels}\n"
        c += f"NAtoms: {self.natoms}\n"
        c += "Cell vectors (Angstrom):\n"
        for vec in self.cell:
            c += f"  {vec[0]:%12.7f} {vec[1]:%12.7f} {vec[2]:%12.7f}\n"
        return c

    # write len function
    def __len__(self):
        return len(self._atoms)

    def sort(self,
             keep_first_order=True) -> List[int]:
        """Classify atoms according to their atom types.
        The new atom type order is the order of first appearance in the original atom list.

        Returns:
            List[int]: Indices of the rearranged atoms in the original list.
        """
        new_atom_list, index = AbacusATOM.sort(self._atoms, keep_first_order=keep_first_order)
        self._atoms = new_atom_list
        return index
    
    @property
    def natoms(self):
        # Number of atoms in the structure
        return len(self._atoms)
    
    @property
    def cell(self):
        # Return the cell vectors
        return self._cell
    
    @property
    def atoms(self):
        # Return the list of AbacusATOM objects
        return self._atoms

    @property
    def atomtypes(self):
        """Return a list of AbacusAtomType objects for unique atom types in the structure.
        """
        return [atom.atomtype for atom in self._atoms]

    @property
    def uniq_atomtypes(self):
        """Return a list of unique AbacusAtomType objects for the structure.
        NOTE: It only merge the consecutive same atom types, which means there may be duplicate types in the returned list.
        If you want to get the true unique atom types, please use sort() method first.
        """
        return AbacusATOM.find_uniq_atomtypes(self._atoms)

    @property
    def labels(self):
        # Return the list of labels of all atoms
        return [atom.label for atom in self._atoms]

    @property
    def elements(self):
        # Return the list of elements of all atoms
        return [atom.element for atom in self._atoms]
    
    @property
    def masses(self):
        # Return the list of masses of all atoms
        return [atom.mass for atom in self._atoms]
    
    @property
    def pps(self):
        # Return the list of pseudopotential file names of all atoms
        return [atom.pp for atom in self._atoms]
    
    @property
    def orbs(self):
        # Return the list of orbital file names of all atoms
        return [atom.orb for atom in self._atoms]
    
    @property
    def paws(self):
        # Return the list of PAW file names of all atoms
        return [atom.paw for atom in self._atoms]
    
    @property
    def coords(self):
        # Return the list of coordinates of all atoms
        return [atom.coord for atom in self._atoms]
    
    @property
    def coords_direct(self):
        return Cartesian2Direct(self.coords, self.cell)
    
    @property
    def moves(self):
        # Return the list of move flags of all atoms
        return [atom.move for atom in self._atoms]
    
    @property
    def atom_mags(self) -> List[Union[float,Tuple[float,float,float]]]:
        # return magnetic moments of all atoms, will transfer to same type if needed
        # e.g., if one atom is noncolinear, all will be noncolinear
        # if atom.mag is None, value will be atom.type_mag
        noncolinear = any(atom.noncolinear for atom in self._atoms)
        mags = []
        for atom in self._atoms:
            if noncolinear and not atom.noncolinear:
                mags.append( (0.0, 0.0, atom.type_mag) )
            else:
                mags.append(atom.atommag)
        return mags

    @cell.setter
    def cell(self, value: Union[List[Tuple[float,float,float]], np.ndarray]):
        if isinstance(value, np.ndarray):
            assert value.ndim == 2 and value.shape == (3,3), "Cell numpy array must be two-dimensional with shape (3,3)."
            value = [tuple(value[i].tolist()) for i in range(3)]
        else:
            assert len(value) == 3 and all(len(vec) == 3 for vec in value), "Cell must be a list of three 3D vectors."
        self._cell = value
    
    @coords.setter
    def coords(self, value: List[Tuple[float,float,float]]):
        assert len(value) == self.natoms, "Number of coordinates must match number of atoms."
        for i in range(self.natoms):
            self._atoms[i].coord = value[i]
    
    @coords_direct.setter
    def coords_direct(self, value: List[Tuple[float,float,float]]):
        assert len(value) == self.natoms, "Number of coordinates must match number of atoms."
        cart_coords = Direct2Cartesian(value, self.cell)
        for i in range(self.natoms):
            self._atoms[i].coord = cart_coords[i]

    def set_pp(self, pp_dict: Dict[str, str], key_type:Literal["element","label"]="element"):
        """Set pseudopotential file names for atoms based on a provided dictionary.

        Args:
            pp_dict (Dict[str, str]): Dictionary mapping element symbols to pseudopotential file names.
        """
        for atom in self._atoms:
            key = atom.element if key_type == "element" else atom.label
            if key in pp_dict:
                atom.pp = pp_dict[key]
            else:
                raise KeyError(f"Pseudopotential for {key_type}: {key} not found in provided dictionary.")

    def set_orb(self, orb_dict: Dict[str, str], key_type: Literal["element","label"]="element"):
        """Set orbital file names for atoms based on a provided dictionary.

        Args:
            orb_dict (Dict[str, str]): Dictionary mapping element symbols to orbital file names.
        """
        for atom in self._atoms:
            key = atom.element if key_type == "element" else atom.label
            if key in orb_dict:
                atom.orb = orb_dict[key]
            else:
                raise KeyError(f"Orbital for {key_type}: {key} not found in provided dictionary.")
    
    def set_paw(self, paw_dict: Dict[str, str], key_type: Literal["element","label"]="element"):
        """Set PAW file names for atoms based on a provided dictionary.

        Args:
            paw_dict (Dict[str, str]): Dictionary mapping element symbols to PAW file names.
        """
        for atom in self._atoms:
            key = atom.element if key_type == "element" else atom.label
            if key in paw_dict:
                atom.paw = paw_dict[key]
            else:
                raise KeyError(f"PAW for {key_type}: {key} not found in provided dictionary.")
    
    def set_coords(self, coords: List[Tuple[float,float,float]], direct: bool=False):
        """
        Set coordinates of all atoms using given coordinates.
        Args:
            coords (List[Tuple[float,float,float]]): List of coordinates for each atom. If direct is False, the unit of coordinates is Angstrom.
            direct (bool): Whether the coordinates are in direct or cartesian coordinates. Default is False.
        """
        if direct:
            coords = Direct2Cartesian(coords, self.cell)
        for i in range(len(self._atoms)):
            self._atoms[i].coord = coords[i]
    
    def pp_dict(self):
        """
        Get dict of pps
        """
        pps = {}
        for atom in self._atoms:
            if atom.label not in pps.keys():
                pps[atom.label] = atom.pp
        return pps
    
    def orb_dict(self):
        """
        Get dict of orbs
        """
        orbs = {}
        for atom in self._atoms:
            if atom.label not in orbs.keys():
                orbs[atom.label] = atom.orb
        return orbs
    
    def paw_dict(self):
        """
        Get dict od paws
        """
        paws = {}
        for atom in self._atoms:
            if atom.label not in paws.keys():
                paws[atom.label] = atom.paw
        return paws

    @staticmethod
    def read(filename: str, fmt: Literal["stru", "abacus/stru", "poscar","vasp", "cif"]="stru") -> "AbacusSTRU":
        """Read structure from a file in the specified format.

        Args:
            filename (str): Input file name.
            fmt (str): Format of the input file. Options are "stru", "poscar", "cif". Default is "stru".

        Returns:
            AbacusSTRU: An AbacusSTRU object representing the structure.
        """
        fmt = fmt.lower()
        if fmt in ["stru", "abacus/stru"]:
            stru_data = read_stru_file(stru=filename)
            cell = (np.array(stru_data["cell"]) * stru_data['lattice_constant'] * BOHR2A).tolist()
            if stru_data["cartesian"]:
                coords = (np.array(stru_data["coord"]) * stru_data['lattice_constant'] * BOHR2A).tolist()
            else:
                coords = Direct2Cartesian(stru_data["coord"], cell)
            atom_list = []
            label_tot = get_total_property(stru_data, "label")
            pp_tot = get_total_property(stru_data, "pp")
            orb_tot = get_total_property(stru_data, "orb")
            paw_tot = get_total_property(stru_data, "paw")
            type_mag_tot = get_total_property(stru_data, "magmom")

            for i in range(len(coords)):
                atom = AbacusATOM(
                    label=label_tot[i],
                    coord=tuple(coords[i]),
                    element=None,
                    mass=None,
                    pp=None if len(stru_data['pp']) == 0 else pp_tot[i],
                    orb=None if len(stru_data['orb']) == 0 else orb_tot[i],
                    paw=None if len(stru_data['paw']) == 0 else paw_tot[i],
                    type_mag=type_mag_tot[i],
                    move=stru_data["move"][i],
                    mag=stru_data["magmom_atom"][i],
                    angle1=stru_data["angle1"][i],
                    angle2=stru_data["angle2"][i],
                    velocity=stru_data["velocity"][i],
                    constrain=stru_data["constrain"][i],
                    lambda_=stru_data["lambda_"][i],
                )
                atom_list.append(atom)
            dpks = stru_data.get("dpks", None)
            metadata = {
                "lattice_constant": stru_data.get("lattice_constant", 1.0),
                "atom_type": "cartesian" if stru_data.get("cartesian", True) else "direct",
            }
            return AbacusSTRU(cell=cell, atoms=atom_list, dpks=dpks, metadata=metadata)
        elif fmt in ["poscar", "vasp", "cif"]:
            # use ase to read poscar/vasp/cif file
            from ase.io import read as ase_read
            if fmt == "poscar":
                fmt = "vasp"
            atom_type = "direct"
            if fmt == "vasp":
                with open(filename, 'r') as f: lines = f.readlines()
                if lines[5].strip().lower().startswith("c"):
                    atom_type = "cartesian"
    
            ase_stru = ase_read(filename, format=fmt)
            return AbacusSTRU.from_ase(ase_stru, metadata={
                "lattice_constant": A2BOHR,
                "atom_type": atom_type,
            })
        else:
            raise ValueError(f"Unsupported format: {fmt}")

    def write(self, filename: str,
              fmt: Literal["stru", "poscar","vasp", "cif"]="stru",
              empty2x: bool = False,
              direct: Optional[bool]=None):
        """Write the structure to a file in the specified format.
        Args:
            filename (str): Output file name.
            fmt (str): Format of the output file. Options are "stru", "poscar", "cif". Default is "stru".
            empty2x (bool): If True, convert 'empty' atoms to 'X' element before writing. Default is False.
            direct (bool, optional): If True, write atomic positions in direct coordinates. If False, write in cartesian coordinates. If None, use the value from metadata. Default is None.
        """
        fmt = fmt.lower()
        if direct is None:
            direct = (self.metadata.get("atom_type","cartesian").lower() == "direct")
        else:
            direct = direct
        
        elements = copy.deepcopy(self.elements)
        if empty2x:
            for i in range(len(self.labels)):
                if "empty" in self.labels[i]:
                    elements[i] = 'X'


        if fmt in ["stru", "abacus/stru"]:
            atom_list = copy.deepcopy(self._atoms)
            unique_types = AbacusATOM.find_uniq_atomtypes(atom_list)
            lc = self.metadata.get("lattice_constant", 1.0)
            cell = np.array(self.cell) * A2BOHR / lc
            coord = np.array([atom.coord for atom in atom_list]) * A2BOHR / lc
            if direct:
                coord = Cartesian2Direct(coord.tolist(), self.cell)
            else:
                coord = coord.tolist()
            cell = cell.tolist()

            write_stru_file(cell=cell, coord=coord, 
                            label=[ut.label for ut in unique_types],
                            atom_number=[ut.natom for ut in unique_types],
                            struf=filename,
                            direct=direct,
                            pp =[ut.pp for ut in unique_types],
                            mass = [ut.mass for ut in unique_types],
                            orb = [ut.orb for ut in unique_types],
                            paw = [ut.paw for ut in unique_types],
                            magmom_global=[ut.type_mag for ut in unique_types],
                            lattice_constant=lc,
                            move = [atom.move for atom in  atom_list],
                            magmom = [atom.mag for atom in atom_list],
                            velocity = [atom.velocity for atom in atom_list],
                            angle1 = [atom.angle1 for atom in atom_list],
                            angle2 = [atom.angle2 for atom in atom_list],
                            constrain = [atom.constrain for atom in atom_list],
                            lambda_ = [atom.lambda_ for atom in atom_list],
                            dpks = self.dpks)
        elif fmt in  ["poscar", "vasp"]:
            write_poscar(cell = self.cell,
                         coord=self.coords if not direct else self.coords_direct,
                         label=elements, poscar=filename, direct=direct, move=self.moves)
        elif fmt == "cif":
            ase_stru = self.to(fmt="ase", empty2x=empty2x)
            ase_stru.write(filename, format="cif")
        else:
            raise ValueError(f"Unsupported format: {fmt}")
    
    @staticmethod
    def from_ase(ase_stru,
                 metadata: Dict[str, Any] = {}) -> "AbacusSTRU":
        """Create an AbacusSTRU object from an ASE Atoms object.
        pps, orbs and paws in the obtained AbacusSTRU object will be from ase_stru.info, which should be a dictionary with keys "pp", "orb", "paw".
        ase_stru.info['pp'] should be a dictionary with keys being element symbols and values being the pp file. "orb" abd "paw" should also be given as "pp".
        Args:
            ase_stru: ASE Atoms object.
            metadata (Dict[str, Any], optional): Additional metadata for the AbacusSTRU object. Default is an empty dictionary.
        Returns:
            AbacusSTRU: An AbacusSTRU object representing the structure.
        """
        from ase import Atoms
        assert isinstance(ase_stru, Atoms), "Input structure must be an ASE Atoms object."
        cell = ase_stru.get_cell().tolist()
        atom_list = []
        mags = ase_stru.arrays["initial_magmoms"] if "initial_magmoms" in ase_stru.arrays.keys() else [None]*len(ase_stru)
        for i in range(len(ase_stru)):
            atom_move = (True, True, True)
            # Get atom coordinate fix from constraints
            for constraint in ase_stru.constraints:
                const_type = type(constraint).__name__
                if const_type == "FixAtoms":
                    if i in constraint.index:
                        atom_move = (False, False, False)
                elif const_type in ["FixCartesian", "FixScaled"]:
                    if i in constraint.index:
                        atom_move = constraint.mask

            atom = AbacusATOM(
                label=ase_stru[i].symbol,
                coord=tuple(ase_stru[i].position.tolist()),
                element=ase_stru[i].symbol,
                mass=ase_stru[i].mass,
                pp=ase_stru.info.get("pp", {}).get(ase_stru[i].symbol, None),
                orb=ase_stru.info.get("orb", {}).get(ase_stru[i].symbol, None),
                paw=ase_stru.info.get("paw", {}).get(ase_stru[i].symbol, None),
                type_mag=0.0,
                move=atom_move,
                mag= mags[i],
            )
            atom_list.append(atom)

        return AbacusSTRU(cell=cell, atoms=atom_list, metadata=metadata)

    @staticmethod
    def from_pymatgen(pymatgen_stru,
        meta_data: Dict[str, Any] = {}) -> "AbacusSTRU":
        """Create an AbacusSTRU object from a Pymatgen Structure object.
        Args:
            pymatgen_stru: Pymatgen Structure object.
            meta_data (Dict[str, Any], optional): Additional metadata for the AbacusSTRU object. Default is an empty dictionary.
        Returns:
            AbacusSTRU: An AbacusSTRU object representing the structure.
        """
        from pymatgen.core import Structure
        assert isinstance(pymatgen_stru, Structure), "Input structure must be a Pymatgen Structure object."
        cell = pymatgen_stru.lattice.matrix.tolist()
        atom_list = []
        for site in pymatgen_stru.sites:
            atom = AbacusATOM(
                label=site.specie.symbol,
                coord=tuple(site.coords.tolist()),
                element=site.specie.symbol,
                mass=site.specie.atomic_mass,
                type_mag=0.0,
                move=(True, True, True),
            )
            atom_list.append(atom)

        return AbacusSTRU(cell=cell, atom_list=atom_list, meta_data=meta_data)
    
    @staticmethod
    def from_dpdata(dpdata_stru,
                    index : int = 0,
                    meta_data: Dict[str, Any] = {}) -> "AbacusSTRU":
        """Create an AbacusSTRU object from a DPData System object.
        Args:
            dpdata_stru: DPData System object.
            index (int): Index of the configuration in the DPData System to convert. Default is 0.
            meta_data (Dict[str, Any], optional): Additional metadata for the AbacusSTRU object. Default is an empty dictionary.
        Returns:
            AbacusSTRU: An AbacusSTRU object representing the structure.
        """
        from dpdata import System
        assert isinstance(dpdata_stru, System), "Input structure must be a DPData System object."
        data = dpdata_stru.data
        coords = data['coords'][index]
        atom_list = []
        for i in range(dpdata_stru.get_natoms()):
            atom_list.append(AbacusATOM(
                label=data['atom_names'][data["atom_types"][i]],
                coord=tuple(coords[i].tolist()),
                element=data['atom_names'][data["atom_types"][i]],
                mass=data["masses"][data["atom_types"][i]],
                type_mag=0.0,
                move=(True, True, True) if "move" not in data else tuple(data["move"][index][i]),
                mag = None if "spins" not in data else data["spins"][index][i]
            ))
        cell = data['cells'][index].tolist()
        return AbacusSTRU(cell=cell, atom_list=atom_list, meta_data=meta_data)
        
    @staticmethod
    def from_phonopy(phonopy_stru,
                     meta_data: Dict[str, Any] = {}) -> "AbacusSTRU":
        """Create an AbacusSTRU object from a Phonopy Atoms object.
        Args:
            phonopy_stru: Phonopy Atoms object.
            meta_data (Dict[str, Any], optional): Additional metadata for the AbacusSTRU object
        Returns:
            AbacusSTRU: An AbacusSTRU object representing the structure.
        """
        from phonopy.structure.atoms import PhonopyAtoms
        assert isinstance(phonopy_stru, PhonopyAtoms), "Input structure must be a Phonopy Atoms object."
        cell = phonopy_stru.cell.tolist()
        atom_list = []
        for i in range(len(phonopy_stru)):
            atom = AbacusATOM(
                label=phonopy_stru.get_chemical_symbols()[i],
                coord=tuple(phonopy_stru.get_positions()[i].tolist()),
                element=phonopy_stru.get_chemical_symbols()[i],
                mass=phonopy_stru.get_masses()[i],
                type_mag=0.0,
                move=(True, True, True),
            )
            atom_list.append(atom)
        return AbacusSTRU(cell=cell, atom_list=atom_list, meta_data=meta_data)
        

    def to(self, fmt: Literal["ase", "pymatgen", "dpdata", "phonopy"],
           empty2x: bool = False):
        """Convert the structure to another format.

        Args:
            fmt (str): Target format. Options are:
             - "ase": ASE Atoms object.
             - "pymatgen": Pymatgen Structure object.
             - "dpdata": DPData System object.
             - "phonopy": Phonopy Atoms object.
            empty2x (bool): If True, convert 'empty' atoms to 'X' element. Default is False.

        Returns:
            Converted structure object.
        """
        elements = self.elements
        if empty2x:
            labels = self.labels
            for i in range(len(labels)):
                if 'empty' in labels[i]:
                    elements[i] = 'X'

        if fmt == "ase":
            from ase import Atoms
            return Atoms(symbols=elements, positions=self.coords, cell=self.cell, pbc=True, magmoms=self.atom_mags,
                          info={"pp": self.pp_dict(),"orb": self.orb_dict(),"paw": self.paw_dict(),"dpks": self.dpks})
        elif fmt == "pymatgen":
            from pymatgen.core import Structure
            return Structure(lattice=self.cell, species=elements, coords=self.coords, coords_are_cartesian=True,labels=self.labels,)
        elif fmt == "dpdata":
            raise NotImplementedError("DPData format conversion is not implemented yet.")
        elif fmt == "phonopy":
            from phonopy.structure.atoms import PhonopyAtoms
            return PhonopyAtoms(symbols=elements, positions=self.coords, cell=self.cell, masses=self.masses,
                            magnetic_moments=self.atom_mags)
        else:
            raise ValueError(f"Unsupported format: {fmt}")
    
    def rotate(self, rot_mat):
        # rotate the cell using given rotation matric
        self.cell = np.dot(rot_mat, np.array(self.cell).T).T.tolist()
        for i in range(len(self._atoms)):
            self._atoms[i].rotate(rot_mat)

    def fix_atom_by_index(self, indices: List[int],
                          move: Optional[Tuple[bool, bool, bool]] = (False, False, False),
                          only: Optional[bool]=False):
        """
        Fix atoms by index.
        
        Args:
            indices (List[int]): List of indices of atoms to fix. Starts from 0.
            move (Tuple[bool, bool, bool]): Tuple indicating whether the atom is allowed to move in each direction. Default is (False, False, False), which means all 3 directions are fixed.
            only (bool): If True, override the move settings for unselected atoms and allow all unselected atoms to move. Default is False.
        """
        for i in range(self.natoms):
            if i in indices:
                self._atoms[i].move = move
            elif only:
                self._atoms[i].move = (True, True, True)

    def fix_atom_by_coord(self,
                          min: float,
                          max: float,
                          cartesian: Optional[bool]=True,
                          direction: Optional[Literal[0, 1, 2]]=2, 
                          move: Optional[Tuple[bool, bool, bool]]=(False, False, False),
                          only: Optional[bool]=False):
        """
        Fix atoms by coordinates (cartesian or direct).
        
        Args:
            min (float): Minimum height of atoms to fix.
            max (float): Maximum height of atoms to fix.
            cartesian (bool): Whether to use cartesian coordinates. Default is True.
            direction (int): Direction of atoms to fix. Can be 0, 1 or 2, means 'x', 'y' and 'z' respectively. Default is 2.
            move: Tuple[bool, bool, bool]: Tuple indicating whether the selected atoms are allowed to move in each direction. Default is (False, False, False), which means all 3 directions are fixed.
            only (bool): If True, override the move settings for unselected atoms and allow all unselected atoms to move. Default is False.
        """
        for i in range(self.natoms):
            if cartesian:
                if self.coords[i][direction] >= min and self.coords[i][direction] <= max:
                    self._atoms[i].move = move
                elif only:
                    self._atoms[i].move = (True, True, True)
            else:
                if self.coords_direct[i][direction] >= min and self.coords_direct[i][direction] <= max:
                    self._atoms[i].move = move
                elif only:
                    self._atoms[i].move = (True, True, True)
    
    def permute_lat_vec(self, mode=Literal["bca", "cab"], rotate_cart_coord=False):
        """
        Permute the axis of the structure.
        Args:
            mode (str): The mode of permuting axis. Can be "bca" or "cab", which means transform from (\vec{a}, \vec{b}, \vec{c}).T to (\vec{b}, \vec{c}, \vec{a}).T or (\vec{c}, \vec{a}, \vec{b}).T).
        """
        if mode == "bca":
            trans_mat = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]])
        elif mode == "cab":
            trans_mat = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]])
        else:
            raise ValueError(f"Invalid mode: {mode}")
        
        self.cell = np.dot(trans_mat, self.cell)

        # rotate cartesian coordinates
        if rotate_cart_coord:
            self.rotate(trans_mat)

        return trans_mat

def parse_stru_position(pos_line):
    '''
  The content in atom position block:
  - `m` or NO key word: three numbers, which take value in 0 or 1, control how the atom move in geometry relaxation calculations. In example below, the numbers `0 0 0` following the coordinates of the first atom means this atom are *not allowed* to move in all three directions, and the numbers `1 1 1` following the coordinates of the second atom means this atom *can* move in all three directions.
  - `v` or `vel` or `velocity`: set the three components of initial velocity of atoms in geometry relaxation calculations(e. g. `v 1.0 1.0 1.0`).
  - `mag` or `magmom` : set the start magnetization for each atom. In colinear case only one number should be given. In non-colinear case one have two choice:either set one number for the norm of magnetization here and specify two polar angle later(e. g. see below), or set three number for the xyz commponent of magnetization here (e. g. `mag 0.0 0.0 1.0`). Note that if this parameter is set, the initial magnetic moment setting in the second line will be overrided.
    - `angle1`: in non-colinear case, specify the angle between c-axis and real spin, in angle measure instead of radian measure
    - `angle2`: in non-colinear case, specify angle between a-axis and real spin in projection in ab-plane , in angle measure instead of radian measure

      e.g.:

      ```
      Fe
      1.0
      2
      0.0 0.0 0.0 m 0 0 0 mag 1.0 angle1 90 angle2 0 cs 0 0 0
      0.5 0.5 0.5 m 1 1 1 mag 1.0 angle1 90 angle2 180
      ```
    '''
    sline = pos_line.split()
    pos = [float(i) for i in sline[:3]]
    move = None
    velocity = None
    magmom = None
    angle1 = None
    angle2 = None
    constrain = None
    lambda1 = None
    if len(sline) > 3:
        mag_list = []
        velocity_list = []
        move_list = []
        angle1_list = []
        angle2_list = []
        constrain_list = []
        lambda_list = []
        label = "move"
        for i in range(3,len(sline)):
            # firstly read the label
            if sline[i] == "m":
                label = "move"
                move_list = []
            elif sline[i] in ["v","vel","velocity"]:
                label = "velocity"
                velocity_list = []
            elif sline[i] in ["mag","magmom"]:
                label = "magmom"
                mag_list = []
            elif sline[i] == "angle1":
                label = "angle1"
                angle1_list = []
            elif sline[i] == "angle2":
                label = "angle2"
                angle2_list = []
            elif sline[i] in ["constrain","sc"]:
                label = "constrain"
                constrain_list = []
            elif sline[i] in ["lambda"]:
                label = "lambda"
                lambda_list = []
            
            # the read the value to the list    
            elif label == "move":
                move_list.append(int(sline[i]))
            elif label == "velocity":
                velocity_list.append(float(sline[i]))
            elif label == "magmom":
                mag_list.append(float(sline[i]))
            elif label == "angle1":
                angle1_list.append(float(sline[i]))
            elif label == "angle2":
                angle2_list.append(float(sline[i]))
            elif label == "constrain":
                constrain_list.append(bool(int(sline[i])))
            elif label == "lambda":
                lambda_list.append(float(sline[i]))
        if len(move_list) == 3:
            move = move_list
        if len(velocity_list) == 3:
            velocity = velocity_list
        if len(mag_list) in [1,3]:
            magmom = mag_list if len(mag_list) == 3 else mag_list[0]
        if len(angle1_list) == 1:
            angle1 = angle1_list[0]
        if len(angle2_list) == 1:
            angle2 = angle2_list[0]
        if len(constrain_list) == 3:
            constrain = constrain_list
        elif len(constrain_list) == 1:
            constrain = constrain_list[0]
        if len(lambda_list) == 3:
            lambda1 = lambda_list
        elif len(lambda_list) == 1:
            lambda1 = lambda_list[0]
            
            
    return pos,move,velocity,magmom,angle1,angle2,constrain,lambda1

def read_stru_file(stru:str = "STRU"):
    '''Read ABACUS STRU file and return a dictionary with structure information.
    Args:
        stru (str): Path to the STRU file. Default is "STRU".
    Returns:
        dict: A dictionary containing structure information with keys:
            - label: list of labels for each atom type
            - atom_number: list of atom numbers for each atom type
            - cell: 3x3 list of cell vectors
            - coords: list of coordinates for each atom
            - pp: list of pseudopotential files for each atom type
            - orb: list of orbital files for each atom type
            - paw: list of PAW files for each atom type
            - lattice_constant: lattice constant value
            - move: list of move flags for each atom
            - magmom_global: list of global magnetic moments for each atom type
            - magmom: list of magnetic moments for each atom
            - velocity: list of velocities for each atom
            - angle1: list of angle1 values for each atom
            - angle2: list of angle2 values for each atom
            - constrain: list of constraints for each atom
            - lambda1: list of lambda values for each atom
    
    NOTE:
        1. Do not support bravais lattice now.
        2. the value for cell/coords/lattice_constant are the exact values in STRU file
        3. Only direct/cartessian coordinate type is supported.
    
    '''
    def get_block(keyname):
        block = []
        for i,line in enumerate(lines):
            if line.strip() == "": continue
            elif line.split('#')[0].strip() == keyname:
                for ij in range(i+1,len(lines)):
                    if lines[ij].strip() == "" or \
                        lines[ij].strip()[0] in ["#"] or\
                        ("//" in lines[ij] and lines[ij].strip()[:2] in ["//"]): continue
                    elif lines[ij].strip() in ABACUS_STRU_KEY_WORD:
                        return block
                    else:
                        block.append(lines[ij].split("#")[0].split("//")[0].strip())
                return block
        return None
    
    if not os.path.isfile(stru):
        return None
    with open(stru) as f1: lines = f1.readlines()  
    atomic_species = get_block("ATOMIC_SPECIES")
    numerical_orbital = get_block("NUMERICAL_ORBITAL")
    lattice_constant = get_block("LATTICE_CONSTANT")
    lattice_vector = get_block("LATTICE_VECTORS")
    atom_positions = get_block("ATOMIC_POSITIONS")
    dpks = get_block("NUMERICAL_DESCRIPTOR")
    pawf = get_block("PAW_FILES")
    lattice_constant = 1.0 if lattice_constant == None else float(lattice_constant[0].split()[0]) 
    dpks = None if dpks == None else dpks[0].strip()
    
    #read species
    pp = []
    labels = []
    mass = []
    for line in atomic_species:
        sline = line.split()
        labels.append(sline[0])
        mass.append(float(sline[1]))
        if len(sline) > 2: 
            pp.append(sline[2])
    if len(pp) == 0:
        pp = None
        
    #read orbital
    if numerical_orbital == None:
        orb = None
    else:
        orb = []
        for line in numerical_orbital:
            orb.append(line.split()[0])
    
    # read paw files
    if pawf == None:
        paw = None
    else:
        paw = []
        for line in pawf:
            paw.append(line.split("#")[0].strip())
    
    #read cell
    cell = []
    try:
        for line in lattice_vector:
            cell.append([float(i) for i in line.split()[:3]])
    except:
        traceback.print_exc()
        print("WARNING: LATTICE_VECTORS is incorrect !!!!!!")
    #read coordinate and coordinate type and atom number of each type
    atom_number = []
    coords = []
    magmom_global = [] # the initial magmom of each type
    magmom = [] # the initial magmom of each atom
    move = []
    velocity = []
    angle1 = []
    angle2 = []
    constrain = [] # the constrain of each atom
    lambda1 = [] # the lambda for delta spin
    coord_type = atom_positions[0].split("#")[0].strip().lower()
    if coord_type.startswith("dire"):
        cartesian = False
    elif coord_type.startswith("cart"):
        cartesian = True
    else:
        print("Not support coordinate type %s now." % atom_positions[0].strip())
        sys.exit(1)
    i = 1
    real_label = []
    real_pp = []
    real_orb = []
    real_paw = []
    while i < len(atom_positions):
        label = atom_positions[i].strip()
        if label not in labels:
            print("label '%s' is not matched that in ATOMIC_SPECIES" % label)
            sys.exit(1)
        an = int(atom_positions[i+2].split()[0])
        if an == 0:
            i += 3
            continue
        
        real_label.append(label)
        label_idx = labels.index(label)
        if pp:
            real_pp.append(pp[label_idx])
        if orb:
            real_orb.append(orb[label_idx])
        if paw:
            real_paw.append(paw[label_idx])
            
        magmom_global.append(float(atom_positions[i+1].split()[0]))
            
        atom_number.append(an)
            
        i += 3
        for j in range(atom_number[-1]):
            pos,imove,ivelocity,imag,iangle1,iangle2,iconstrain,ilambda1 = parse_stru_position(atom_positions[i+j])
            coords.append(pos)
            move.append(imove)
            velocity.append(ivelocity)
            magmom.append(imag)
            angle1.append(iangle1)
            angle2.append(iangle2)
            constrain.append(iconstrain)
            lambda1.append(ilambda1)
            
        i += atom_number[-1]

    return {
        "label": real_label,   # list of labels for each atom type
        "atom_number": atom_number,  # list of atom numbers for each atom type
        "cell": cell,              # 3x3 list of cell vectors
        "coord": coords,       # list of coordinates for each atom
        "pp": real_pp,          # list of pseudopotential files for each atom type
        "orb": real_orb,        # list of orbital files for each atom type
        "paw": real_paw,        # list of PAW files for each atom type
        "lattice_constant": lattice_constant,  # lattice constant value
        "move": move,              # list of move flags for each atom
        "magmom": magmom_global, # list of global magnetic moments for each atom type
        "magmom_atom": magmom,       # list of magnetic moments for each atom
        "velocity": velocity,    # list of velocities for each atom
        "angle1": angle1,       # list of angle1 values for each atom
        "angle2": angle2,       # list of angle2 values for each atom
        "constrain": constrain, # list of constrain flags for each atom
        "lambda_": lambda1,     # list of lambda1 values for each atom
        "dpks": dpks,           # dpks value
        "cartesian": cartesian  # boolean indicating if coordinates are cartesian
    }



def write_stru_file(
          cell:List[Tuple[float,float,float]],
          coord: List[Tuple[float,float,float]],
          label:List[str],
          atom_number:List[int],
          struf: str ="STRU", 
          direct:bool=False,
          pp:Optional[List[str]] = None,
          mass:Optional[List[float]] = None,
          orb:Optional[List[str]] = None,
          paw:Optional[List[str]] = None,
          magmom_global:Optional[List[float]] = None,
          lattice_constant:float=1.0,
          move:Optional[List[Optional[Tuple[bool,bool,bool]]]] = None,
          magmom:Optional[List[float]] = None,
          velocity:Optional[List[Optional[Tuple[float,float,float]]]] = None,
          angle1:Optional[List[Optional[float]]] = None,
          angle2:Optional[List[Optional[float]]] = None,
          constrain:Optional[List[Optional[Union[bool,Tuple[bool,bool,bool]]]]] = None,
          lambda_:Optional[List[Optional[Union[float,Tuple[float,float,float]]]]] = None,
          dpks:Optional[str] = None,
          ):
    '''Write to ABACUS STRU file. 
    1. The value will be directly write to STRU file, no unit conversion.
       So the unit of cell/coord/lattice_constant should be bohr, and the coord should be in direct/cartesian type according to `direct` parameter.
       The real cell should be cell * lattice_constant.
    2. The order of pp/mass/orb/paw/magmom_global should be the same as label list.
    3. The coord should be in the order of label and atom_number.
    4. The order of coord/move/magmom/velocity/angle1/angle2/constrain/lambda_ should be the same as coord list.
    '''
    # check parameters
    assert len(label) == len(atom_number), "label and atom_number length mismatch"
    natoms = sum(atom_number)
    assert len(coord) == natoms, "coord length mismatch with atom_number"
    for key, lst in [("pp",pp),("mass",mass),("magmom_global",magmom_global),
                     ("orb",orb),("paw",paw)]:
        if lst is not None:
            assert len(lst) == len(label), f"{key} length mismatch with label"
    for key, lst in [("move",move),("magmom",magmom),("velocity",velocity),
                     ("angle1",angle1),("angle2",angle2),("constrain",constrain),("lambda_",lambda_)]:
        if lst is not None:
            assert len(lst) == natoms, f"{key} length mismatch with natoms"
    
    cc = ""
    #write species
    cc += "ATOMIC_SPECIES\n"
    for i, ilabel in enumerate(label):
        cc += f"{ilabel} "
        if mass and mass[i] is not None:
            cc += f"{mass[i]} "
        else:
            cc += "1.0 "
        if pp and pp[i] is not None:
            cc += f"{pp[i]}\n"
        else:
            cc += "\n"
    
    if orb and None not in orb:
        #write orb
        cc += "\nNUMERICAL_ORBITAL\n"
        for i in orb:
            cc += i + "\n"
    elif all([i is None for i in orb]):
        pass
    elif orb and None in orb:
        print("WARNING: orb list contains None, skip writing NUMERICAL_ORBITAL block")
    
    if paw and None not in paw:
        # write pawfile
        cc += "\nPAW_FILES\n"
        for i in paw:
            cc += i + "\n"
    elif all([i is None for i in paw]):
        pass
    elif paw and None in paw:
        print("WARNING: paw list contains None, skip writing PAW_FILES block")
    
    # write LATTICE_CONSTANT
    cc += "\nLATTICE_CONSTANT\n%f\n" % lattice_constant
    #write LATTICE_VECTORS
    cc += "\nLATTICE_VECTORS\n"
    for i in cell:
        cc += "%17.11f %17.11f %17.11f\n" % tuple(i)
    #write ATOMIC_POSITIONS
    cc += "\nATOMIC_POSITIONS\n"
    if not direct:
        cc += "Cartesian\n"
    else:
        cc += "Direct\n"
    icoord = 0
    for i,ilabel in enumerate(label):
        cc += "\n%s\n" % ilabel
        if magmom_global and magmom_global[i] is not None:
            cc += "%f\n" % magmom_global[i]
        else:
            cc += "0.0\n"
        cc += "%d\n" % atom_number[i]
        for j in range(atom_number[i]):
            cc += "%17.11f %17.11f %17.11f " % tuple(coord[icoord + j])
            if move and move[icoord + j] and len(move[icoord + j]) == 3:
                cc += "%d %d %d " % tuple(move[icoord + j])
            if velocity and velocity[icoord + j] and len(velocity[icoord + j]) == 3:
                cc += "v %f %f %f " % tuple(velocity[icoord + j])
            if magmom and magmom[icoord + j] is not None:
                if isinstance(magmom[icoord + j],list):
                    if len(magmom[icoord + j]) == 3:
                        cc += "mag %12.8f %12.8f %12.8f " % tuple(magmom[icoord + j])
                    elif len(magmom[icoord + j]) == 1:
                        cc += "mag %12.8f " % magmom[icoord + j][0]
                elif magmom[icoord + j] != None:
                    cc += "mag %12.8f " % magmom[icoord + j]
            if angle1 and angle1[icoord + j] != None:
                    cc += "angle1 %f " % angle1[icoord + j]
            if angle2 and angle2[icoord + j] != None:
                    cc += "angle2 %f " % angle2[icoord + j]
            if constrain and constrain[icoord + j]:
                if isinstance(constrain[icoord + j],list) and len(constrain[icoord + j]) == 3:
                    cc += "sc " + " ".join(["1" if ic else "0" for ic in constrain[icoord + j]]) + " " 
                elif isinstance(constrain[icoord + j],list) and len(constrain[icoord + j]) == 1:
                    cc += "sc " + ("1" if constrain[icoord + j][0] else "0") + " "
                elif not isinstance(constrain[icoord + j],list):
                    cc += "sc " + ("1" if constrain[icoord + j] else "0") + " "
                else:
                    print("ERROR: the constrain is not a list or a bool value, skip it")
                    print("\t\tconstrain:",constrain[icoord + j])
            if lambda_ and lambda_[icoord + j]:
                if isinstance(lambda_[icoord + j],list):
                    cc += "lambda " + " ".join([str(ic) for ic in lambda_[icoord + j]]) + " "
                else:
                    cc += "lambda " + str(lambda_[icoord + j]) + " "
            cc += "\n"
        icoord += atom_number[i]

    if dpks:
        cc += "\nNUMERICAL_DESCRIPTOR\n"
        cc += dpks    

    os.makedirs(os.path.dirname(os.path.abspath(struf)), exist_ok=True)
    with open(struf,"w") as f1:
        f1.write(cc)
    
    return cc

def write_poscar(
          cell:List[Tuple[float,float,float]],
          coord: List[Tuple[float,float,float]],
          label:List[str],
          poscar: str ="POSCAR", 
          direct:bool=True,
          move:Optional[List[Optional[Tuple[bool,bool,bool]]]] = None,
          ):
    '''Write to VASP POSCAR file. 
    
    Args:
        cell (List[Tuple[float,float,float]]): 3x3 list of cell vectors in Angstrom.
        coord (List[Tuple[float,float,float]]): List of coordinates for each atom in Angstrom (if direct=False) or in direct coordinates (if direct=True).
        label (List[str]): List of labels for each atom.
        poscar (str): Path to the POSCAR file. Default is "POSCAR".
        direct (bool): If True, write atomic positions in direct coordinates. If False, write in cartesian coordinates. Default is True.
        move (Optional[List[Optional[Tuple[bool,bool,bool]]]]): List of move flags for each atom. Default is None.
    '''
    # check parameters
    assert len(label) == len(coord), "label and coord length mismatch"
    if move:
        assert len(move) == len(coord), "move length mismatch with coord"
    
    # find unique labels and their counts
    unique_labels = [label[0]]
    atom_number = []
    natoms = 1
    for i in range(1,len(label)):
        if label[i] != label[i-1]:
            atom_number.append(natoms)
            unique_labels.append(label[i])
            natoms = 1
        else:
            natoms += 1
    atom_number.append(natoms) 

    cc = ""
    #write header
    cc += "poscar\n"
    cc += "1.0\n"
    #write cell
    for i in cell:
        cc += "%17.11f %17.11f %17.11f\n" % tuple(i)
    #write atom types and numbers
    cc += " ".join(unique_labels) + "\n"
    cc += " ".join([str(num) for num in atom_number]) + "\n"
    #write coordinates
    if direct:
        cc += "Direct\n"
    else:
        cc += "Cartesian\n"
    for i in range(len(coord)):
        cc += "%17.11f %17.11f %17.11f" % tuple(coord[i])
        if move and move[i] and len(move[i]) == 3:
            cc += " " + " ".join(["T" if mv else "F" for mv in move[i]])
        cc += "\n"
    os.makedirs(os.path.dirname(poscar), exist_ok=True)
    with open(poscar,"w") as f1:
        f1.write(cc)
    return cc

def get_total_property(stru_data: Dict[str, Any],
                       prop: Literal['label', 'magmom', 'pp', 'orb', 'paw']):
    """
    Get selected property in stru_data (read by read_stru_file) for each atom stored for each type .
    Args:
        stru_data (Dict[str, Any]): The structure data.
        prop (Literal['label', 'magmom', 'pp', 'orb', 'paw']): The property name.
    Returns:
        float: The total property.
    """
    result = []
    if len(stru_data[prop]) == 0:
        return []
    else:
        assert len(stru_data[prop]) == len(stru_data['atom_number'])
        for item, count in zip(stru_data[prop], stru_data['atom_number']):
            result.extend([item] * count)

        return result



import os,sys,traceback
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element

HARTREE2EV = 27.211396132
EV2RY = 2.0 / HARTREE2EV
RY2EV = HARTREE2EV / 2.0
BOHR2A = 0.52917721092
KBAR2HARTREEPERBOHR3 = 3.398927420868445E-6
KBAR2EVPERANGSTROM3 = KBAR2HARTREEPERBOHR3 * HARTREE2EV / BOHR2A**3

def ReadFile(ifile,warn=True):
    if ifile == None:
        return []

    if os.path.isfile(ifile):
        with open(ifile) as f1:
            context = f1.readlines()
    else:
        if warn: print("WARNING: can not find file %s" % ifile)
        context = []

    return context

def FindOutput(path,keyinfo):
    'find the file that has keyinfo'
    cwd = os.getcwd()
    os.chdir(path)
    allfiles = os.listdir(".")
    for ifile in allfiles:
        if os.path.isfile(ifile):
            find = False
            try:
                with open(ifile) as f1: lines = f1.read()
                if keyinfo in lines:
                    find = True
            except:
                pass
                
            if find:
                os.chdir(cwd)
                return os.path.join(path,ifile)
    os.chdir(cwd)
    return None

def ReadXmlFile(ifile,warn=True):
    if ifile == None:
        return None

    if os.path.isfile(ifile):
        try:
            tree = ET.parse(ifile)
            return tree.getroot()
        except:
            traceback.print_exc()
            if warn: print("WARNING: can not parse file %s" % ifile)
            return None
    else:
        if warn: print("WARNING: can not find file %s" % ifile)
        return None

def XmlFindMultiLayer(root,layerlist):
    tmp = root
    for i in layerlist:
        if tmp == None: return None
        tmp = tmp.find(i)
    return tmp

def XmlFindMultiLayerText(root,layerlist):
    tmp = XmlFindMultiLayer(root,layerlist)
    if isinstance(tmp,Element):
        return tmp.text
    return None

def XmlGetText(tmp1,func=lambda x:x,idx: int=None):
    funcname = sys._getframe().f_code.co_name
    if tmp1 == None:
        return None
    elif isinstance(tmp1,list):
        if idx == None:
            return [func(i.text) for i in tmp1]
        else:
            try:
                return func(tmp1[idx].text)
            except:
                print("ERROR: %s, try to get the text of %d-th element, the length of list is %d" 
                        % (funcname,idx,len(tmp1)))
                traceback.print_exc()
                return None
    elif isinstance(tmp1,Element):
        return func(tmp1.text)
    else:
        print("Not support type '%s' in function '%s' now." % 
              (type(tmp1),funcname))
        return None

def istr(x,n=None):
    if type(x) == float and n != None:
            return "%%.%df" % n % x
    return str(x)

def iint(x):
    try:
        return int(x)
    except:
        return None

def ifloat(x):
    try:
        return float(x)
    except:
        return None

def ibool(x):
    try:
        if x.lower() == "true":
            return True
        elif x.lower() == "false":
            return False
        else:
            return bool(int(x))
    except:
        return None

def imath(a,b,symbol):
    if a == None or b == None:
        return None
    elif symbol == '*':
        return a*b
    elif symbol == '+':
        return a+b
    elif symbol == '-':
        return a-b
    elif symbol == '/':
        return a/b
    else:
        print("Not support '%s' now" % symbol)
        return None

def strtime2sec(stim_str):
    # time_str: 1d2h3m4s
    
    day = 0
    hour = 0
    minute = 0
    second = 0
    time_str = stim_str.lower()
    if 'd' in time_str:
        day = int(time_str.split('d')[0])
        time_str = time_str.split('d')[1]
    if 'h' in time_str:
        hour = int(time_str.split('h')[0])
        time_str = time_str.split('h')[1]
    if 'm' in time_str:
        minute = int(time_str.split('m')[0])
        time_str = time_str.split('m')[1]
    if 's' in time_str:
        second = float(time_str.split('s')[0])
    return day*24*3600 + hour*3600 + minute*60 + second

def cal_band_gap(band,efermi):
    # band: [[spin1],[spin2],...]
    # efermi: [efermi_spin1,efermi_spin2]
    if isinstance(efermi,float):
        efermi = [efermi,efermi]
            
    cb = None
    vb = None
    
    for ispin in range(len(band)):
        nband_below_fermi = None
        for ik in range(len(band[ispin])):
            for ib in range(len(band[ispin][ik])):
                if band[ispin][ik][ib] > efermi[ispin]:
                    if nband_below_fermi == None:
                        nband_below_fermi = ib
                    elif nband_below_fermi != ib:
                        return 0
                    icb = band[ispin][ik][ib-1] - efermi[ispin]
                    ivb = band[ispin][ik][ib] - efermi[ispin]
                    if cb == None or icb > cb:
                        cb = icb
                    if vb == None or ivb < vb:
                        vb = ivb
                    break
    
    if cb == None or vb == None:
        band_gap = None
    else:
        band_gap = vb - cb
        if band_gap < 0: band_gap = 0 
    return band_gap

def plot_band(band,fname,efermi=None):
    """
    band: NSPIN X NKPTS X NBAND
    fname: the file name of the band plot
    efermi: float or [float, float]
    """
    import numpy as np
    import matplotlib.pyplot as plt
    
    def plot_one(ax, iband, efermi=None, _range=None):
        for j in range(len(iband)):
            ax.plot(iband[j], "--o", linewidth=0.3, markersize=0.3)
        if efermi is not None:
            ax.axhline(y=efermi, color='r', linestyle='--', linewidth=0.3)
            ax.text(int(len(iband[0])/3), efermi, f'Efermi={efermi:.3f} eV', color='r', fontsize=8)
        if _range is not None:
            ax.set_ylim(_range)
        else:
            ax.set_ylim(np.min(iband.flatten()) - 1, np.max(iband.flatten()) + 1)
        ax.set_ylabel('Energy (eV)')
        ax.set_xlabel('k point')
        ax.set_xlim(-1, len(iband[0]))
        
    band = np.array(band)
    nband = len(band)
    if isinstance(efermi, (float, type(None),int)):
        efermi = [efermi] * nband

    fig, axs = plt.subplots(nband, 2, figsize=(8, 4*nband))
    axs = axs.flatten()
    for i in range(nband):
        plot_one(axs[2*i], np.array(band[i]).T, efermi=efermi[i])
        _range = [-5, 5] if efermi is None else [efermi[i]-5, efermi[i]+5]
        plot_one(axs[2*i+1], np.array(band[i]).T, efermi=efermi[i], _range=_range)
        axs[2*i].set_title(f'Spin {i+1} (Full)')
        axs[2*i+1].set_title(f'Spin {i+1} (Zoomed)')
        
    fig.tight_layout()
    plt.savefig(fname, dpi=300)
    plt.close()

def abacus_orb_label(l,m):
    if l == 0:
        return "s"
    elif l == 1:
        if m == 0:
            return "p_z"
        elif m == 1:
            return "p_x"
        elif m == 2:
            return "p_y"
    elif l == 2:
        if m == 0:
            return "d_z^2"
        elif m == 1:
            return "d_xz"
        elif m == 2:
            return "d_yz"
        elif m == 3:
            return "d_x^2-y^2"
        elif m == 4:
            return "d_xy"

    return f"{l}_{m}"

def get_abacus_json(json_dict,key_list):
    # need first find if key_list in json_dict
    v = None
    has_keys = True
    for k in key_list:
        if k in json_dict:
            json_dict = json_dict[k]
        else:
            has_keys = False
            break
    if has_keys:
        v = json_dict
    return v,has_keys

def get_metric_from_str(istr):
    # istr: "{energy}/{natom}"
    # the metric name may in the format of "{metric_name}"
    # return a list of metric name
    
    if "{" not in istr:
        return [istr]

    metric_list = []
    for i in istr.split("{"):
        if "}" in i:
            metric_list.append(i.split("}")[0])
    return metric_list

def get_mulliken(mullikenf):
    """Read mulliken population analysis file
    
    Return:
        atom_mag: list of list, atom_mag[step][atom] = mag or [mx,my,mz]
        atom_elec: list of list, atom_elec[step][atom] = [elec_spin1,(elec_spin2)]
    """
    atom_mag = []
    atom_elec = []
    atom_labels = []
    with open(mullikenf) as f1: lines = f1.readlines()
    for idx,line in enumerate(lines):
        if line[:5] == "STEP:":
            atom_mag.append([])
            atom_elec.append([])
            atom_labels.append([])
        elif "Total Magnetism on atom" in line:
            if "(" in line and ")" in line:
                # for nspin = 4
                sline = line.split("(")[1].split(")")[0].split(",")
                if len(sline) == 3:
                    atom_mag[-1].append([float(sline[0]),float(sline[1]),float(sline[2])])
            else:
                atom_mag[-1].append(float(line.split()[-1]))
        elif "Zeta of" in line:
            atom_labels[-1].append(line.split()[3])
            two_spin = True if "Spin 2" in line else False
            atom_elec[-1].append([])
            j = idx + 1
            while j < len(lines) :
                if "Zeta of" in lines[j]:
                    break
                if "sum over m+zeta" in lines[j] or "SUM OVER M+Zeta" in lines[j] and "M+Zeta+L" not in lines[j]:
                    atom_elec[-1][-1].append([float(lines[j].split()[3])])
                    if two_spin:
                        atom_elec[-1][-1][-1].append(float(lines[j].split()[4]))
                j += 1
    return atom_mag,atom_elec, atom_labels


import numpy as np

def cal_norm_vector(v1, v2):
    """
    Calculate the normal vector of two vectors
    
    Parameters
    ----------
    v1 : list of three floats
        The first vector
    v2 : list of three floats
        The second vector
        
    Returns
    -------
    normal_v : list of three floats
        The normal vector of v1 and v2
    """
    v1 = np.array(v1) / np.linalg.norm(v1)
    v2 = np.array(v2) / np.linalg.norm(v2)
    dot = np.dot(v1, v2)
    if abs(dot - 1) < 1e-6:
        # the two vectors are parallel, then we need to find a vector that is not parallel to v1
        for i in [[1,0,0], [0,1,0], [0,0,1]]:
            if abs(np.dot(v1, i) - 1) > 1e-6:
                normal_v = np.cross(v1, i)
                return normal_v / np.linalg.norm(normal_v)
    else:
        normal_v = np.cross(v1, v2)
        return normal_v / np.linalg.norm(normal_v)

def cal_angle(v1, v2):
    """
    Calculate the angle between two vectors
    
    Parameters
    ----------
    v1 : list of three floats
        The first vector
    v2 : list of three floats
        The second vector
        
    Returns
    -------
    angle : float
        The angle between v1 and v2 in degree
    """
    v1 = np.array(v1) / np.linalg.norm(v1)
    v2 = np.array(v2) / np.linalg.norm(v2)
    dot = np.dot(v1, v2)
    return np.arccos(dot) * 180 / np.pi

def prepare_atom_mag(mag1, mag2, step, number, together_half=True):
    """
    Prepare the magnetic moment of the atoms
    
    Parameters
    ----------
    mag1 : list of three floats
        The magnetic moment of the first atom
    mag2 : list of three floats
        The magnetic moment of the second atom
    step : float
        The step size of the tilting angle
    number : int
        The number of tilting angles
    together_half : bool
        Whether the tilting angle of tilting atom1 and atom2 together is half of the tilting angle of tilting atom1 and tilting atom2 separately    
    
        
    Returns
    -------
    mag : list of lists
        The dimension is number * 3 * 2 * 3.
        Like: [[[mag1, mag2],[mag1, mag2], [mag1, mag2]],...]
        For each tilting angle, will return three pairs of magnetic moments, which are tilting atom1, tilting atom2 and tilting atom1 and atom2 together. 
    """
    
    mags = []
    norm1 = np.linalg.norm(mag1)
    norm2 = np.linalg.norm(mag2)
    v1 = mag1 / norm1
    v2 = mag2 / norm2
    angle0 = cal_angle(v1, v2) # the angle between mag1 and mag2
    normal_v = cal_norm_vector(v1, v2) # the normal vector of mag1 and mag2
    normal_v1 = cal_norm_vector(v1, normal_v) # the normal vector of mag1 and normal_v
    normal_v2 = cal_norm_vector(normal_v, v2) # the normal vector of mag2 and normal_v
    
    coef_together = 0.5 if together_half else 1
    
    for i in range(number):
        d_angle = (i+1) * step
        angle1 = angle0 + d_angle
        angle2 = angle0 + d_angle * coef_together
        mag = []
        # case1: tilting atom1
        mag1_tilt = v2 * np.cos(angle1 * np.pi / 180) * norm1 + normal_v2 * np.sin(angle1 * np.pi / 180) * norm1
        mag.append([mag1_tilt, mag2])
        
        # case2: tilting atom2
        mag2_tilt = v1 * np.cos(angle1 * np.pi / 180) * norm2 + normal_v1 * np.sin(angle1 * np.pi / 180) * norm2
        mag.append([mag1, mag2_tilt])
        
        # case3: tilting atom1 and atom2 together
        mag1_tilt = v2 * np.cos(angle2 * np.pi / 180) * norm1 + normal_v2 * np.sin(angle2 * np.pi / 180) * norm1
        mag2_tilt = v1 * np.cos(angle2 * np.pi / 180) * norm2 + normal_v1 * np.sin(angle2 * np.pi / 180) * norm2
        mag.append([mag1_tilt, mag2_tilt])
        
        mags.append(mag)
    
    if True:
        # check the result
        print("Original magnetic moments:")
        print("Atom1:", mag1)
        print("Atom2:", mag2)
        print("The angle between the two magnetic moments:", angle0)
        print("The normal vector of the two magnetic moments:", normal_v)
        print("The normal vector of mag1 and normal_v:", normal_v1)
        print("The normal vector of mag2 and normal_v:", normal_v2)
        print("The prepared magnetic moments:")
        for i, mag in enumerate(mags):
            for j, mag_pair in enumerate(mag):
                print(f"Angle {i}, Case {j+1}:", "%.3f" % cal_angle(mag_pair[0], mag_pair[1]), mag_pair)
    
    return np.array(mags).tolist()
        





from ..resultAbacus import ResultAbacus

class RefAbacus(ResultAbacus): 
    # This class is used to compare the results of the current calculation with the reference results.  
    
    @ResultAbacus.register(delta_energy="the difference between energy and the reference value. Unit in eV. Key in reference file is \"energy\"",
                           delta_energyPerAtom="delta_energy/natom, unit in eV")
    def GetDeltaEnergy(self): 
        if self.resultREF.get("energy",None) != None:
            if self["energy"] != None:
                self["delta_energy"] = self["energy"] - self.resultREF.get("energy")
                if self["natom"] != None:
                    self["delta_energyPerAtom"] = (self["energy"] - self.resultREF.get("energy")) / self["natom"]
                else:
                    self["delta_energyPerAtom"] = None
                return
        self["delta_energy"] = None
        self["delta_energyPerAtom"] = None
        
    @ResultAbacus.register(delta_force="the difference between force and the reference value. The reference value should be a list of 3*natom, each list is a force of each atom. Unit in eV/Angstrom",
                           delta_stress="the difference between stress and the reference value. The reference value should be a list 9 elements, unit in kbar"
                           )
    def GetDeltaForceStress(self): 
        df = None
        ds = None
        if self.resultREF.get("force",None) != None:
            if self["force"] != None:
                try:
                    df = [i - j for i,j in zip(self["force"],self.resultREF.get("force"))]
                except:
                    df = None
        if self.resultREF.get("stress",None) != None:
            if self["stress"] != None:
                try:
                    ds = [i - j for i,j in zip(self["stress"],self.resultREF.get("stress"))]
                except:
                    ds = None
        self["delta_force"] = df
        self["delta_stress"] = ds


from ..model import Model
import json, os
from typing import List, Dict, Any, Literal, Tuple, Optional, Union
import re
import shutil, glob, copy
from pathlib import Path
from itertools import groupby

import numpy as np
from ase.build import bulk, make_supercell
from ase import Atoms
from ase.io import read, write

from abacustest.lib_prepare.abacus import WriteKpt, WriteInput, ReadInput, AbacusStru, ReadKpt
from abacustest.lib_prepare.comm import kpt2kspacing
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE, ELEMENT_CRYSTAL_STRUCTURES, A2BOHR
from abacustest.lib_collectdata.collectdata import RESULT
from abacustest.lib_model.model_013_inputs import PrepInput, InputsModel
from abacustest.outresult import pandas_out

class VacancyModel(Model):
    @staticmethod
    def model_name(): # type: ignore
        """
        Name of the model, which will be used as the subcommand
        """
        return "vacancy"
    
    @staticmethod
    def description(): # type: ignore
        '''
        Description of the model
        '''
        return "Calculate the vacancy formation energy for uncharged systems"
    
    @staticmethod
    def prepare_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand 
        '''
        parser.description = "Prepare the inputs for vacancy formation energy calculation."
        parser.add_argument('-j', '--job', default=[], action="extend", nargs="*", help='the paths of structure files or ABACUS jobs. If structure files, should be in CIF, VASP POSCAR or ABACUS STRU format. If ABACUS job dirs, should contain INPUT, STRU, or KPT, and pseudopotential and orbital files')
        parser.add_argument('-s', '--supercell', type=int, default=[1, 1, 1], nargs=3, help='the supercell size, default is [1, 1, 1]')
        parser.add_argument('-i', "--index", type=int, default=None, nargs="*", help="Index of the atom to be removed. Start from 1")
        parser.add_argument("--index-file", type=str, default=None, help="The file containing the indices of atoms to be removed for each job. Each line corresponds to a job and the indices are separated by space. If provided, will override the --index argument.")
        parser.add_argument("--cal-reference", type=bool, default=True, help="Whether to do cell-relax calculation for elemental crystals. If not, a file containing element and its energy should be provided.")
        parser.add_argument("--ref-dir", type=str, default='ref_element', help="The directory containing the jobs of elemental crystal structures.")
        parser.add_argument("--max-step", type=int, default=100, help="The maximum number of steps to relax calculation. Default is 100.")
        parser.add_argument("--force-thr-ev", type=float, default=0.01, help="The threshold of force convergence")
        parser.add_argument("--stress-thr-kbar", type=float, default=0.5, help="The threshold of stress convergence")
        parser.add_argument("--image", type=str, default=RECOMMAND_IMAGE, help="The image to use for the Bohrium job, default is %s" % RECOMMAND_IMAGE)
        parser.add_argument("--machine", type=str, default=RECOMMAND_MACHINE, help="The machine to use for the Bohrium job, default is 'c32_m64_cpu'.")
        parser.add_argument("--abacus_command", type=str, default=RECOMMAND_COMMAND, help=f"The command to run the Abacus job, default is '{RECOMMAND_COMMAND}'.")
        parser.add_argument('--ftype', type=str, default='cif', help='The format of the structure files. Can be "cif", "poscar" or "stru". Default is cif.')
        parser.add_argument("--pp",default=None,type=str,help="the path of pseudopotential library, or read from enviroment variable ABACUS_PP_PATH")
        parser.add_argument("--orb",default=None,type=str,help="the path of orbital library, or read from enviroment variable ABACUS_ORB_PATH")
        parser.add_argument("--input",default=None,type=str,help="the template of input file, if not specified, the default input will be generated")
        parser.add_argument("--relax-kspacing", default=None, type=float, help="the kspacing for kpoint generation of relax calculation, default is None, which means kspacing in original input file will be used.")
        parser.add_argument("--scf-kspacing", default=None, type=float, help="the kspacing for scf calculation after relax, default is None, which means kspacing in original input file will be used.")
        parser.add_argument("--lcao", action="store_true", help="whether to use lcao basis, default is pw basis")
        parser.add_argument("--nspin", default=1, type=int, choices=[1, 2, 4], help="the number of spins, can be 1 (no spin), 2 (spin polarized), or 4 (non-collinear spin). Default is 1.")
        parser.add_argument("--soc", action="store_true", help="whether to use spin-orbit coupling, if True, nspin should be 4.")
        parser.add_argument("--dftu", action="store_true", help="whether to use DFT+U, default is False.")
        parser.add_argument("--dftu_param", default=None, nargs="+", help="the DFT+U parameters, should be element symbol and U value pairs like 'Fe 4 Ti 1'. If dftu is set, but dftu_param is not set, the default U values will be used: 4 eV for d orbital elements and 6 eV for f orbital elements.")
        parser.add_argument("--init_mag", default=None, nargs="+", help="the initial magnetic moment for magnetic elements, should be element symbol and magnetic moment pairs like 'Fe 4 Ti 1'.")
        parser.add_argument("--afm", action="store_true", help="whether to use antiferromagnetic calculation, default is False. Only valid when init_mag is set.")
        parser.add_argument("--copy_pp_orb", action="store_true", help="whether to copy the pseudopotential and orbital files to each job directory or link them. Default is False, which means linking the files.")

        return parser
    
    def run_prepare(self, params):
        '''
        Run the model with the given parameters
        '''
        if not params.job:
            raise ValueError("No job specified, please use -j or --job to specify the job paths.")
        if params.dftu_param is not None:
            dftu_param = InputsModel.parse_dftu_param(params.dftu_param)
        else:
            dftu_param = None
            
        if params.init_mag is not None:
            init_mag = InputsModel.parse_init_mag(params.init_mag)
        else:
            init_mag = None
        
        if params.index_file is None and params.index is None:
            raise ValueError("No vacancy index specified, please use --index or --index-file to specify the atom indices to be removed.")
        
        folders = prepare_vacancy_jobs(jobs=params.job,
                                       supercell=params.supercell,
                                       vacancy_indices=params.index_file if params.index_file is not None else params.index,
                                       cal_reference=params.cal_reference,
                                       ref_dir=params.ref_dir,
                                       max_step=params.max_step,
                                       force_thr_ev=params.force_thr_ev,
                                       stress_thr_kbar=params.stress_thr_kbar,
                                       ftype=params.ftype,
                                       pp=params.pp,
                                       orb=params.orb,
                                       input=params.input,
                                       relax_kspacing=params.relax_kspacing,
                                       scf_kspacing=params.scf_kspacing,
                                       lcao=params.lcao,
                                       nspin=params.nspin,
                                       soc=params.soc,
                                       dftu=params.dftu,
                                       dftu_param=dftu_param,
                                       init_mag=init_mag,
                                       afm=params.afm,
                                       copy_pp_orb=params.copy_pp_orb)
        
        run_sh = f"""{params.abacus_command} | tee log
if [ -d "final_scf" ]; then
    cp OUT.*/STRU_ION_D final_scf/STRU
    cd final_scf
    {params.abacus_command} | tee log
    cd ..
fi
"""
        # Write run scripts
        with open("run.sh", "w") as f1:
            f1.write(run_sh)
        
        setting = {
            "save_path": "results",
            "run_dft": [{
                "ifrun": True,
                "example": folders,
                "extra_files":["run.sh"],
                "command": "bash run.sh",
                "image": params.image,
                "bohrium": {
                    "scass_type": params.machine,
                    "job_type": "container",
                    "platform": "ali"
                }
            }]
        } 
        
        setting_file = "setting.json"
        json.dump(setting, open(setting_file, "w"), indent=4)
        print("\nThe inputs are generated in", ", ".join(params.job))
        print(f"You can modify '{setting_file}' and 'run.sh', and execute below command to run the abacustest to submit all jobs to bohrium:\n\tabacustest submit -p setting.json\n")
        print(f"After finishing the calculations, you can enter the results directory, \nand run below command below to postprocess the vacancy formation energy results:\n\tabacustest model vacancy post -j ...\n")
        

    def postprocess_args(parser):
        '''
        Add arguments for the postprocess subcommand
        The arguments can not be command, model, modelcommand 
        '''
        parser.description = "Postprocess the vacancy formation energy calculation results."
        parser.add_argument('-j', '--job', default=[], action="extend", nargs="*", help='the paths of the job directories, should contain the results of deformed structures generated by the prepare step.')
        parser.add_argument('--ref-dir', default='ref_element', help="The ABACUS job directory containing the elemental crystal structures. If not provided, reference atom energies will be read from provided reference file.")
        parser.add_argument("--ref-file", type=str, default='ref_energy.txt', help="The file to store the atom energy reference. Will be generated after a vacancy formation energy calculation." + \
                            "Has two columns: atom label, and atom energy. If not provided or selected atom does not contained in the atom energy reference file, the atom energy will be automatically calculated.")
        return parser   

    def run_postprocess(self, params):
        '''
        Parse the parameters and run the postprocess process
        '''
        metrics, ref_atom_energies = postprocess_vacancy(params.job, params.ref_dir, params.ref_file)
        write_ref_atom_energies(params.ref_file, ref_atom_energies)
        json.dump(metrics, open("metrics_vacancy.json", "w"), indent=4)
        pandas_out(metrics)
        print(f"\nThe vacancy formation results are saved in 'metrics_vacancy.json'\nThe reference atom energies are saved in '{params.ref_file}'\n")

def read_ref_atom_energies(ref_atom_energy_file: str) -> Dict[str, float]:
    '''
    Read the reference atom energies from the file.
    '''
    ref_atom_energies = {}
    with open(ref_atom_energy_file, "r") as f:
        for line in f:
            label, energy = line.strip().split()
            ref_atom_energies[label] = float(energy)
    return ref_atom_energies

def write_ref_atom_energies(ref_atom_energy_file: str, ref_atom_energies: Dict[str, float]):
    '''
    Write the reference atom energies to the file.
    '''
    with open(ref_atom_energy_file, "w") as f:
        for label, energy in ref_atom_energies.items():
            f.write(f"{label} {energy}\n")

def parser_indices_file(index_file: str) -> Dict[str, List[int]]:
    """
    Parse the index file to get the vacancy indices for each job.
    The file should contain multiple lines, each line corresponds to a job and contains the indices of atoms to be removed, separated by space.
    Returns a dictionary with job names as keys and list of indices as values.

    Config file example:
    job1 1 2 3
    job2 4 5 6
    """
    job_indices = {}
    with open(index_file, "r") as f:
        for line in f:
            if line.strip() == "":
                continue
            parts = line.strip().split()
            job_name = parts[0]
            indices = [int(idx) for idx in parts[1:]]
            assert all(idx > 0 for idx in indices), f"All indices should be positive integers in line: {line.strip()}"
            job_indices[job_name] = indices
    return job_indices

def prepare_vacancy_jobs(
    jobs: str,
    supercell: Tuple[int, int, int],
    vacancy_indices: Union[str, List[int]],
    cal_reference: bool = True,
    ref_dir: str = "ref_element",
    max_step: int = 100,
    force_thr_ev: float = 0.01,
    stress_thr_kbar: float = 0.5,
    ftype: str = "cif",
    pp: Optional[str] = None,
    orb: Optional[str] = None,
    input: Optional[str] = None,
    relax_kspacing: Optional[float] = None,
    scf_kspacing: Optional[float] = None,
    lcao: bool = False,
    nspin: int = 1,
    soc: bool = False,
    dftu: bool = False,
    dftu_param: Optional[Dict[str, Union[float, Tuple[Literal["s", 'p', "d"], float]]]] = None,
    init_mag: Optional[Dict[str, float]] = None,
    afm: bool = False,
    copy_pp_orb: bool = False
) -> None:
    '''
    Prepare the vacancy calculation jobs.
    '''
    #ref_atom_energies = read_ref_atom_energies(ref_file)
    if pp is None:
        pp = os.environ.get("ABACUS_PP_PATH", None)
    if orb is None:
        orb = os.environ.get("ABACUS_ORB_PATH", None)
    prepared_elements = []
    if cal_reference:
        os.makedirs(ref_dir, exist_ok=True)

    if isinstance(vacancy_indices, list):
        job_indices_dict = {job: vacancy_indices for job in jobs}
    else:
        try:
            job_indices_dict = parser_indices_file(vacancy_indices)
        except Exception as e:
            raise ValueError(f"Error parsing vacancy indices file: {e}")

    folders = []
    for job in jobs:
        print("Job: ", job)
        if job not in job_indices_dict:
            raise ValueError(f"Vacancy indices not specified for job {job} in index file.")
        original_vacancy_indices = remove_redundant_indices(job_indices_dict[job])
        real_vacancy_indices = copy.deepcopy(original_vacancy_indices)
        if not os.path.isdir(job):
            # Create structure with vacancy
            #print((job, ftype, "scf", pp, orb, input, kpt, lcao, nspin, soc, dftu, dftu_param, init_mag, afm, copy_pp_orb))
            setting, job_path = PrepInput(files=job, 
                                          filetype=ftype, 
                                          jobtype="scf", 
                                          pp_path=pp, 
                                          orb_path=orb, 
                                          input_file=input, 
                                          kpt=None, 
                                          lcao=lcao, 
                                          nspin=nspin,
                                          soc=soc,
                                          dftu=dftu,
                                          dftu_param=dftu_param,
                                          init_mag=init_mag,
                                          afm=afm,
                                          copy_pp_orb=copy_pp_orb).run()
            
            if ftype != "stru":
                # Transform the vacancy indices to those in ABACUS STRU file
                categorized_idx = get_categorized_idx(job, ftype)
                for i in range(len(original_vacancy_indices)):
                    # Get global index in ABACUS STRU file
                    assert original_vacancy_indices[i]-1 in categorized_idx, f"Vacancy index {original_vacancy_indices[i]} is out of range for job {job}."
                    real_vacancy_indices[i] = categorized_idx.index(original_vacancy_indices[i]-1) + 1 # Use atom index starting from 1
            
            new_path = job.lstrip("./").replace("/", "_") + "_VACANCY"
            if os.path.exists(new_path):
                shutil.rmtree(new_path)
            shutil.move(job_path[0], new_path)

            job = new_path
            

        print("Preparing vacancy formation energy calculation for job path:", job)
        # clean old folders
        old_folders = [f for f in glob.glob(os.path.join(job, "vacancy_*")) if os.path.isdir(f)]
        if len(old_folders) > 0:
            print(f"Remove old vacancy_* folders in job path: {job}")
            for f in old_folders:
                shutil.rmtree(f)
        
        if not os.path.isfile(os.path.join(job, "STRU")):
            raise FileNotFoundError(f"STRU file not found in job path: {job}")
        if not os.path.isfile(os.path.join(job, "INPUT")):
            raise FileNotFoundError(f"INPUT file not found in job path: {job}")
        
        input_params = ReadInput(os.path.join(job, "INPUT"))
        original_stru = AbacusStru.ReadStru(os.path.join(job, input_params.get('stru_file', 'STRU')))

        # all indices should be valid
        for idx in real_vacancy_indices:
            if idx < 1 or idx > original_stru.get_natoms():
                raise ValueError(f"Vacancy index {idx} is out of range for job {job} with {original_stru.get_natoms()} atoms.")

        input_params["suffix"] = "ABACUS"
        input_params['calculation'] = 'cell-relax'
        input_params['relax_method'] = 'cg'
        input_params['symmetry'] = 0
        input_params['stru_file'] = None
        input_params['ntype'] = None
        input_params['relax_nmax'] = max_step
        input_params['force_thr_ev'] = force_thr_ev
        input_params['stress_thr'] = stress_thr_kbar
        if relax_kspacing is not None:
            input_params['kspacing'] = relax_kspacing

        pp_orb_files_fullname = glob.glob(os.path.join(job, "*.upf")) + glob.glob(os.path.join(job, "*.UPF")) \
                       + glob.glob(os.path.join(job, "*.orb"))
        
        pp_orb_files = [os.path.basename(i) for i in pp_orb_files_fullname]
        original_kpt_file = os.path.join(job, input_params.get('kpt_file', 'KPT'))

        # Prepare input files for the supercell
        # cell-relax job for the original structure
        original_stru_jobpath = os.path.join(job, "vacancy_original_stru")
        os.makedirs(original_stru_jobpath, exist_ok=True)
        create_new_abacus_job(job,
                              original_stru_jobpath,
                              pp_orb_files,
                              extra_input=input_params,
                              stru=original_stru, 
                              kpt_file=original_kpt_file)
        # scf job after cell-relax job for the original structure
        if scf_kspacing is None:
            if 'kspacing' not in input_params.keys():
                original_kpt = ReadKpt(os.path.join(original_stru_jobpath, "KPT"))
                original_kspacing = kpt2kspacing(original_kpt[0][:3], original_stru.get_cell())
            else:
                original_kspacing = input_params['kspacing']
            scf_kspacing = min(original_kspacing*0.9, 0.10)
        
        create_new_abacus_job(job,
                              os.path.join(original_stru_jobpath, "final_scf"),
                              pp_orb_files,
                              extra_input={'calculation': 'scf', 'kspacing': scf_kspacing},
                              stru=None, # Don't prepare STRU file
                              kpt_file=None)
        
        folders.append(original_stru_jobpath)

        # Prepare input files for the supercell with defect
        for i, idx in enumerate(real_vacancy_indices):
            # If ABACUS inputs directory is provided
            vacancy_element, labelidx = original_stru.globalidx2labelidx(idx-1) # Use atom index staring from 0 in globalidx2labelidx
            supercell_stru = original_stru.supercell(supercell)
            defect_supercell_stru = copy.deepcopy(supercell_stru)
            defect_supercell_stru.set_empty_atom(vacancy_element, labelidx)
            defect_supercell_jobpath = os.path.join(job, f"vacancy_defect_{original_vacancy_indices[i]}_{vacancy_element}_{supercell[0]}_{supercell[1]}_{supercell[2]}")
            os.makedirs(defect_supercell_jobpath, exist_ok=True)
            # cell-relax job for the defect structure
            create_new_abacus_job(job,
                                  defect_supercell_jobpath,
                                  pp_orb_files,
                                  extra_input=input_params,
                                  stru=defect_supercell_stru,
                                  kpt_file=original_kpt_file)
            # scf job after cell-relax job for the defect structure
            create_new_abacus_job(job,
                                  os.path.join(defect_supercell_jobpath, "final_scf"),
                                  pp_orb_files,
                                  extra_input={'calculation': 'scf', 'kspacing': scf_kspacing},
                                  stru=None, # Don't prepare STRU file
                                  kpt_file=None) 
            defect_supercell_stru.write2cif(os.path.join(defect_supercell_jobpath, "STRU.cif"), empty2x=True)
            folders.append(defect_supercell_jobpath)

            # Prepare input files for the crystal structure of the vacancy element
            if cal_reference and vacancy_element not in prepared_elements:
                element_type_index = [key for key, group in groupby(original_stru.get_element(number=False))].index(vacancy_element)
                vacancy_element_crys_jobpath = os.path.join(ref_dir, f"{vacancy_element}")
                # Remove old folders when cal_reference is True
                if os.path.isdir(vacancy_element_crys_jobpath):
                    shutil.rmtree(vacancy_element_crys_jobpath)
                vacancy_element_pp = original_stru.get_pp()[element_type_index]
                vacancy_element_orb = original_stru.get_orb()[element_type_index]
                vacancy_element_crys_stru = build_most_stable_elementary_crys_stru(vacancy_element, vacancy_element_pp, vacancy_element_orb)

                create_new_abacus_job(job,
                                      vacancy_element_crys_jobpath,
                                      pp_orb_files,
                                      extra_input=input_params,
                                      stru=vacancy_element_crys_stru,
                                      kpt_file=None)
                
                folders.append(vacancy_element_crys_jobpath)
                prepared_elements.append(vacancy_element)
    
    return folders

def remove_redundant_indices(indices: List[int]):
    """
    Remove the same atom indices in the list.
    """
    new_indices = []
    for ind in indices:
        if ind not in new_indices:
            new_indices.append(ind)
        else:
            print(f"The atom index {ind} is already in the list, skip")
    
    return new_indices

def postprocess_vacancy(jobs: List[str],
                        ref_dir: str,
                        ref_file: str = None) -> Dict[str, Any]:
    '''
    Postprocess the vacancy calculation results.
    
    Args:
        jobs: List of job paths.
    
    Returns:
        A dictionary containing the vacancy results.
    '''
    # Get reference atom energies from file
    if ref_file is None or not os.path.isfile(ref_file):
        ref_atom_energies = {}
    else:
        ref_atom_energies = read_ref_atom_energies(ref_file)
    
    # Get element crystal results from calculation results. Will overwrite the reference atom energies from file
    if ref_dir is not None and os.path.isdir(ref_dir):
        for element_crys_job in glob.glob(os.path.join(ref_dir, "*")):
            vac_ele_crys_stru = AbacusStru.ReadStru(os.path.join(element_crys_job, "STRU"))
            vacancy_element = vac_ele_crys_stru.get_element(number=False)[0]
            element_crys_job_result = read_metrics(element_crys_job, jobtype="relax")
            e_vac_elem_crys = element_crys_job_result["energies"][-1]
            ref_atom_energies[vacancy_element] = e_vac_elem_crys / vac_ele_crys_stru.get_natoms()

    metrics = {}
    for job in jobs:
        print("Postprocessing vacancy calculation for job path:", job)

        sub_folders = [f for f in glob.glob(os.path.join(job, "vacancy_*")) if os.path.isdir(f)]

        defect_supercell_job_results, defect_supercell_scf_results = {}, {}
        for sub_folder in sub_folders:
            if os.path.basename(sub_folder).startswith("vacancy_original_stru"):
                original_stru_job_results = read_metrics(sub_folder, jobtype="relax")
                original_stru_scf_results = read_metrics(os.path.join(sub_folder, 'final_scf'), jobtype="scf")
            if os.path.basename(sub_folder).startswith("vacancy_defect"):
                words = sub_folder.split('/')[-1].split("_")
                vacancy_element, idx = words[3], int(words[2])
                supercell = (int(words[4]), int(words[5]), int(words[6]))
                defect_supercell_job_results[f'{vacancy_element}{idx}'] = read_metrics(sub_folder, jobtype="relax")
                defect_supercell_scf_results[f'{vacancy_element}{idx}'] = read_metrics(os.path.join(sub_folder, 'final_scf'), jobtype="scf")

        for site in defect_supercell_job_results.keys():
            if original_stru_job_results["energies"] is None or defect_supercell_job_results[site]["energies"] is None:
                e_vac_form = None
            elif original_stru_scf_results["converge"] is False or defect_supercell_scf_results[site]["converge"] is False:
                e_vac_form = None
            else:
                e_vac_form = (defect_supercell_scf_results[site]["energy"] + ref_atom_energies[vacancy_element]) - original_stru_scf_results["energy"] * supercell[0] * supercell[1] * supercell[2]

            results = {
                'vac_formation_energy': e_vac_form,
                'original_stru_job_relax_converge': original_stru_job_results['relax_converge'],
                'original_stru_job_normal_end': original_stru_job_results['normal_end'],
                'original_stru_job_max_force': original_stru_job_results['largest_gradient'][-1],
                'original_stru_job_max_stress': original_stru_job_results['largest_gradient_stress'][-1],
                'original_stru_relaxed_lattice_constant': original_stru_job_results['lattice_constant'],
                'defect_supercell_job_relax_converge': defect_supercell_job_results[site]['relax_converge'],
                'defect_supercell_job_normal_end': defect_supercell_job_results[site]['normal_end'],
                'defect_supercell_job_max_force': defect_supercell_job_results[site]['largest_gradient'][-1],
                'defect_supercell_job_max_stress': defect_supercell_job_results[site]['largest_gradient_stress'][-1],
                'defect_supercell_relaxed_lattice_constant': defect_supercell_job_results[site]['lattice_constant'],
            }
            metrics[f"{job.rstrip('/')}-{site}"] = results

    return metrics, ref_atom_energies

def create_new_abacus_job(src_dir: str,
                          dst_dir: str,
                          pp_orb_files: str,
                          extra_input: dict=None,
                          stru: Tuple[AbacusStru, str]=None,
                          kpt_file: str=None):
    """
    Create new abacus job from a source directory and new settings.
    """
    if not os.path.isdir(dst_dir):
        os.makedirs(dst_dir, exist_ok=True)
    
    for file in pp_orb_files:
        os.symlink(os.path.abspath(os.path.join(src_dir, file)), os.path.join(dst_dir, file))
    
    if extra_input is None:
        shutil.copy(os.path.join(src_dir, "INPUT"), os.path.join(dst_dir, "INPUT"))
    else:
        input_params = ReadInput(os.path.join(src_dir, "INPUT"))
        for key, value in extra_input.items():
            input_params[key] = value
        WriteInput(input_params, os.path.join(dst_dir, "INPUT"))
    
    
    if stru is not None:
        if isinstance(stru, AbacusStru):
            stru.write(os.path.join(dst_dir, "STRU"))
        elif isinstance(stru, str):
            shutil.copy(stru, os.path.join(dst_dir, "STRU"))

    if kpt_file is not None and os.path.isfile(kpt_file):
        shutil.copy(kpt_file, os.path.join(dst_dir, os.path.basename(kpt_file)))

def build_most_stable_elementary_crys_stru(element: str, pp: str, orb: str) -> AbacusStru:
    """
    Build the most stable crystal structure of an element.
    Args:
        element (str): The chemical symbol of the element.
        pp (dict): Pseudopotential information from the original structure.
        orb (dict): Orbital information from the original structure.
    Returns:
        AbacusStru: The most stable crystal structure of the element.
    Raises:
        ValueError: If the element is not supported.
    """
    if element not in ELEMENT_CRYSTAL_STRUCTURES:
        raise ValueError(f"Element {element} not supported.")
    else:
        crystal_structure = ELEMENT_CRYSTAL_STRUCTURES[element]
        if crystal_structure['crystal'] == 'bcc':
            stru = bulk(element, 'bcc', a=crystal_structure['a'])
        elif crystal_structure['crystal'] == 'fcc':
            stru = bulk(element, 'fcc', a=crystal_structure['a'])
        elif crystal_structure['crystal'] == 'hcp':
            stru = bulk(element, 'hcp', a=crystal_structure['a'], c=crystal_structure['c'])
        elif crystal_structure['crystal'] == 'diamond':
            stru = bulk(element, 'diamond', a=crystal_structure['a'])
        else:
            raise ValueError(f"Crystal structure {crystal_structure['crystal']} not supported.")
    
    stru_abacus = AbacusStru(label=stru.get_chemical_symbols(),
                             cell=stru.get_cell(),
                             coord=stru.get_positions(),
                             lattice_constant=A2BOHR,
                             cartesian=True)
    stru_abacus.set_pp([pp])
    stru_abacus.set_orb([orb])

    return stru_abacus

def read_metrics(job: str, jobtype: Literal['relax', 'scf']) -> List[str]:
    """
    Read the relaxation metrics from the log file.
    Args:
        job (str): The path to the job directory.
    Returns:
        List[str]: The relaxation metrics.
    """
    results = RESULT(path=job, fmt='abacus')
    if jobtype == 'relax':
        relax_metrics = ["normal_end", "relax_steps", "largest_gradient",
                         "largest_gradient_stress", "relax_converge", "energies", "lattice_constant"]
        return {k: results[k] for k in relax_metrics}
    elif jobtype == 'scf':
        scf_metrics = ["normal_end", "energy", "converge"]
        return {k: results[k] for k in scf_metrics}
    else:
        raise ValueError(f"Job type {jobtype} not supported.")

def get_categorized_idx(original_stru_file: str, original_stru_format: str):
    """
    Get index correspondace between original structure file and ABACUS STRU file.
    Return a list, where its index is the index in the transformed ABACUS STRU file, and its value is the index in the original structure file.
    """
    original_stru = read(original_stru_file, format=original_stru_format)

    unique_labels = []
    for label in original_stru.get_chemical_symbols():
        if label not in unique_labels:
            unique_labels.append(label)
    
    original_indices = []
    for label in unique_labels:
        for i, atom_label in enumerate(original_stru.get_chemical_symbols()):
            if atom_label == label:
                original_indices.append(i)
    
    return original_indices
